""" Module for PINN """
import torch
try:
    from torch.optim.lr_scheduler import LRScheduler  # torch >= 2.0
except ImportError:
    from torch.optim.lr_scheduler import _LRScheduler as LRScheduler  # torch < 2.0

from torch.optim.lr_scheduler import ConstantLR

from .solver import SolverInterface
from ..label_tensor import LabelTensor
from ..utils import check_consistency
from ..loss import LossInterface
from ..problem import InverseProblem
from torch.nn.modules.loss import _Loss

torch.pi = torch.acos(torch.zeros(1)).item() * 2  # which is 3.1415927410125732


class PINN(SolverInterface):
    """
    PINN solver class. This class implements Physics Informed Neural
    Network solvers, using a user specified ``model`` to solve a specific
    ``problem``. It can be used for solving both forward and inverse problems.

    .. seealso::

        **Original reference**: Karniadakis, G. E., Kevrekidis, I. G., Lu, L.,
        Perdikaris, P., Wang, S., & Yang, L. (2021).
        Physics-informed machine learning. Nature Reviews Physics, 3(6), 422-440.
        <https://doi.org/10.1038/s42254-021-00314-5>`_.
    """

    def __init__(
        self,
        problem,
        model,
        extra_features=None,
        loss=torch.nn.MSELoss(),
        optimizer=torch.optim.Adam,
        optimizer_kwargs={'lr': 0.001},
        scheduler=ConstantLR,
        scheduler_kwargs={
            "factor": 1,
            "total_iters": 0
        },
    ):
        '''
        :param AbstractProblem problem: The formulation of the problem.
        :param torch.nn.Module model: The neural network model to use.
        :param torch.nn.Module loss: The loss function used as minimizer,
            default :class:`torch.nn.MSELoss`.
        :param torch.nn.Module extra_features: The additional input
            features to use as augmented input.
        :param torch.optim.Optimizer optimizer: The neural network optimizer to
            use; default is :class:`torch.optim.Adam`.
        :param dict optimizer_kwargs: Optimizer constructor keyword args.
        :param torch.optim.LRScheduler scheduler: Learning
            rate scheduler.
        :param dict scheduler_kwargs: LR scheduler constructor keyword args.
        '''
        super().__init__(models=[model],
                         problem=problem,
                         optimizers=[optimizer],
                         optimizers_kwargs=[optimizer_kwargs],
                         extra_features=extra_features)

        # check consistency
        check_consistency(scheduler, LRScheduler, subclass=True)
        check_consistency(scheduler_kwargs, dict)
        check_consistency(loss, (LossInterface, _Loss), subclass=False)

        # assign variables
        self._scheduler = scheduler(self.optimizers[0], **scheduler_kwargs)
        self._loss = loss
        self._neural_net = self.models[0]

        # inverse problem handling
        if isinstance(self.problem, InverseProblem):
            self._params = self.problem.unknown_parameters
        else:
            self._params = None

    def forward(self, x):
        """
        Forward pass implementation for the PINN
        solver.

        :param torch.Tensor x: Input tensor.
        :return: PINN solution.
        :rtype: torch.Tensor
        """
        return self.neural_net(x)

    def configure_optimizers(self):
        """
        Optimizer configuration for the PINN
        solver.

        :return: The optimizers and the schedulers
        :rtype: tuple(list, list)
        """
        # if the problem is an InverseProblem, add the unknown parameters
        # to the parameters that the optimizer needs to optimize
        if isinstance(self.problem, InverseProblem):
            self.optimizers[0].add_param_group(
                {'params': [self._params[var] for var in self.problem.unknown_variables]}
                )
        return self.optimizers, [self.scheduler]

    def _clamp_inverse_problem_params(self):
        for v in self._params:
            self._params[v].data.clamp_(
                    self.problem.unknown_parameter_domain.range_[v][0],
                    self.problem.unknown_parameter_domain.range_[v][1])

    def _loss_data(self, input, output):
        return self.loss(self.forward(input), output)

    def _loss_phys(self, samples, equation):
        try:
            residual = equation.residual(samples, self.forward(samples))
        except TypeError: # this occurs when the function has three inputs, i.e. inverse problem
            residual = equation.residual(samples, self.forward(samples), self._params)
        return self.loss(torch.zeros_like(residual, requires_grad=True), residual)

    def training_step(self, batch, batch_idx):
        """
        PINN solver training step.

        :param batch: The batch element in the dataloader.
        :type batch: tuple
        :param batch_idx: The batch index.
        :type batch_idx: int
        :return: The sum of the loss functions.
        :rtype: LabelTensor
        """

        dataloader = self.trainer.train_dataloader
        condition_losses = []

        condition_idx = batch['condition']

        for condition_id in range(condition_idx.min(), condition_idx.max()+1):

            condition_name = dataloader.condition_names[condition_id]
            condition = self.problem.conditions[condition_name]
            pts = batch['pts']

            if len(batch) == 2:
                samples = pts[condition_idx == condition_id]
                loss = self._loss_phys(samples, condition.equation)
            elif len(batch) == 3:
                samples = pts[condition_idx == condition_id]
                ground_truth = batch['output'][condition_idx == condition_id]
                loss = self._loss_data(samples, ground_truth)
            else:
                raise ValueError("Batch size not supported")

            # TODO for users this us hard to remember when creating a new solver, to fix in a smarter way
            loss = loss.as_subclass(torch.Tensor)

#            # add condition losses and accumulate logging for each epoch
            condition_losses.append(loss * condition.data_weight)
            self.log(condition_name + '_loss', float(loss),
                     prog_bar=True, logger=True, on_epoch=True, on_step=False)

        # clamp unknown parameters of the InverseProblem to their domain ranges (if needed)
        if isinstance(self.problem, InverseProblem):
            self._clamp_inverse_problem_params()

        # TODO Fix the bug, tot_loss is a label tensor without labels
        # we need to pass it as a torch tensor to make everything work
        total_loss = sum(condition_losses)
        self.log('mean_loss', float(total_loss / len(condition_losses)),
                 prog_bar=True, logger=True, on_epoch=True, on_step=False)

        return total_loss

    @property
    def scheduler(self):
        """
        Scheduler for the PINN training.
        """
        return self._scheduler

    @property
    def neural_net(self):
        """
        Neural network for the PINN training.
        """
        return self._neural_net

    @property
    def loss(self):
        """
        Loss for the PINN training.
        """
        return self._loss



__all__ = [
    'SystemEquation',
    'Equation',
    'FixedValue',
    'FixedGradient',
    'FixedFlux',
    'Laplace',
]

from .equation import Equation
from .equation_factory import FixedFlux, FixedGradient, Laplace, FixedValue
from .system_equation import SystemEquation



""" Module for GAROM """

import torch
try:
    from torch.optim.lr_scheduler import LRScheduler  # torch >= 2.0
except ImportError:
    from torch.optim.lr_scheduler import _LRScheduler as LRScheduler  # torch < 2.0

from torch.optim.lr_scheduler import ConstantLR
from .solver import SolverInterface
from ..utils import check_consistency
from ..loss import LossInterface, PowerLoss
from torch.nn.modules.loss import _Loss


class GAROM(SolverInterface):
    """
    GAROM solver class. This class implements Generative Adversarial
    Reduced Order Model solver, using user specified ``models`` to solve
    a specific order reduction``problem``. 

    .. seealso::

        **Original reference**: Coscia, D., Demo, N., & Rozza, G. (2023).
        *Generative Adversarial Reduced Order Modelling*. 
        DOI: `arXiv preprint arXiv:2305.15881.
        <https://doi.org/10.48550/arXiv.2305.15881>`_.
    """

    def __init__(
        self,
        problem,
        generator,
        discriminator,
        loss=None,
        optimizer_generator=torch.optim.Adam,
        optimizer_generator_kwargs={'lr': 0.001},
        optimizer_discriminator=torch.optim.Adam,
        optimizer_discriminator_kwargs={'lr': 0.001},
        scheduler_generator=ConstantLR,
        scheduler_generator_kwargs={
            "factor": 1,
            "total_iters": 0
        },
        scheduler_discriminator=ConstantLR,
        scheduler_discriminator_kwargs={
            "factor": 1,
            "total_iters": 0
        },
        gamma=0.3,
        lambda_k=0.001,
        regularizer=False,
    ):
        """
        :param AbstractProblem problem: The formualation of the problem.
        :param torch.nn.Module generator: The neural network model to use
            for the generator.
        :param torch.nn.Module discriminator: The neural network model to use
            for the discriminator.
        :param torch.nn.Module loss: The loss function used as minimizer,
            default ``None``. If ``loss`` is ``None`` the defualt
            ``PowerLoss(p=1)`` is used, as in the original paper.
        :param torch.optim.Optimizer optimizer_generator: The neural
            network optimizer to use for the generator network
            , default is `torch.optim.Adam`.
        :param dict optimizer_generator_kwargs: Optimizer constructor keyword
            args. for the generator.
        :param torch.optim.Optimizer optimizer_discriminator: The neural
            network optimizer to use for the discriminator network
            , default is `torch.optim.Adam`.
        :param dict optimizer_discriminator_kwargs: Optimizer constructor keyword
            args. for the discriminator.
        :param torch.optim.LRScheduler scheduler_generator: Learning
            rate scheduler for the generator.
        :param dict scheduler_generator_kwargs: LR scheduler constructor keyword args.
        :param torch.optim.LRScheduler scheduler_discriminator: Learning
            rate scheduler for the discriminator.
        :param dict scheduler_discriminator_kwargs: LR scheduler constructor keyword args.
        :param gamma: Ratio of expected loss for generator and discriminator, defaults to 0.3.
        :type gamma: float
        :param lambda_k: Learning rate for control theory optimization, defaults to 0.001.
        :type lambda_k: float
        :param regularizer: Regularization term in the GAROM loss, defaults to False.
        :type regularizer: bool

        .. warning::
            The algorithm works only for data-driven model. Hence in the ``problem`` definition
            the codition must only contain ``input_points`` (e.g. coefficient parameters, time
            parameters), and ``output_points``.
        """

        super().__init__(
            models=[generator, discriminator],
            problem=problem,
            optimizers=[optimizer_generator, optimizer_discriminator],
            optimizers_kwargs=[
                optimizer_generator_kwargs, optimizer_discriminator_kwargs
            ])

        # set automatic optimization for GANs
        self.automatic_optimization = False

        # set loss
        if loss is None:
            loss = PowerLoss(p=1)

        # check consistency
        check_consistency(scheduler_generator, LRScheduler, subclass=True)
        check_consistency(scheduler_generator_kwargs, dict)
        check_consistency(scheduler_discriminator, LRScheduler, subclass=True)
        check_consistency(scheduler_discriminator_kwargs, dict)
        check_consistency(loss, (LossInterface, _Loss))
        check_consistency(gamma, float)
        check_consistency(lambda_k, float)
        check_consistency(regularizer, bool)

        # assign schedulers
        self._schedulers = [
            scheduler_generator(
                self.optimizers[0], **scheduler_generator_kwargs),
            scheduler_discriminator(
                self.optimizers[1],
                **scheduler_discriminator_kwargs)
        ]

        # loss and writer 
        self._loss = loss

        # began hyperparameters
        self.k = 0
        self.gamma = gamma
        self.lambda_k = lambda_k
        self.regularizer = float(regularizer)

    def forward(self, x, mc_steps=20, variance=False):
        """
        Forward step for GAROM solver

        :param x: The input tensor.
        :type x: torch.Tensor
        :param mc_steps: Number of montecarlo samples to approximate the 
            expected value, defaults to 20.
        :type mc_steps: int
        :param variance: Returining also the sample variance of the solution, defaults to False.
        :type variance: bool
        :return: The expected value of the generator distribution. If ``variance=True`` also the
            sample variance is returned.
        :rtype: torch.Tensor | tuple(torch.Tensor, torch.Tensor)
        """

        # sampling
        field_sample = [self.sample(x) for _ in range(mc_steps)]
        field_sample = torch.stack(field_sample)

        # extract mean
        mean = field_sample.mean(dim=0)

        if variance:
            var = field_sample.var(dim=0)
            return mean, var

        return mean

    def configure_optimizers(self):
        """
        Optimizer configuration for the GAROM
        solver.

        :return: The optimizers and the schedulers
        :rtype: tuple(list, list)
        """
        return self.optimizers, self._schedulers

    def sample(self, x):
        # sampling
        return self.generator(x)

    def _train_generator(self, parameters, snapshots):
        """
        Private method to train the generator network.
        """
        optimizer = self.optimizer_generator
        optimizer.zero_grad()

        generated_snapshots = self.generator(parameters)

        # generator loss
        r_loss = self._loss(snapshots, generated_snapshots)
        d_fake = self.discriminator.forward_map([generated_snapshots, parameters])
        g_loss = self._loss(d_fake, generated_snapshots) + self.regularizer * r_loss

        # backward step
        g_loss.backward()
        optimizer.step()

        return r_loss, g_loss

    def _train_discriminator(self, parameters, snapshots):
        """
        Private method to train the discriminator network.
        """
        optimizer = self.optimizer_discriminator
        optimizer.zero_grad()

        # Generate a batch of images
        generated_snapshots = self.generator(parameters)

        # Discriminator pass
        d_real = self.discriminator.forward_map([snapshots, parameters])
        d_fake = self.discriminator.forward_map([generated_snapshots, parameters]) 

        # evaluate loss
        d_loss_real = self._loss(d_real, snapshots)
        d_loss_fake = self._loss(d_fake, generated_snapshots.detach())
        d_loss = d_loss_real - self.k * d_loss_fake

        # backward step
        d_loss.backward(retain_graph=True)
        optimizer.step()

        return d_loss_real, d_loss_fake, d_loss

    def _update_weights(self, d_loss_real, d_loss_fake):
        """
        Private method to Update the weights of the generator and discriminator
        networks.
        """

        diff = torch.mean(self.gamma * d_loss_real - d_loss_fake)

        # Update weight term for fake samples
        self.k += self.lambda_k * diff.item()
        self.k = min(max(self.k, 0), 1)  # Constraint to interval [0, 1]
        return diff
    
    def training_step(self, batch, batch_idx):
        """GAROM solver training step.

        :param batch: The batch element in the dataloader.
        :type batch: tuple
        :param batch_idx: The batch index.
        :type batch_idx: int
        :return: The sum of the loss functions.
        :rtype: LabelTensor
        """

        dataloader = self.trainer.train_dataloader
        condition_idx = batch['condition']

        for condition_id in range(condition_idx.min(), condition_idx.max()+1):

            condition_name = dataloader.condition_names[condition_id]
            condition = self.problem.conditions[condition_name]
            pts = batch['pts']
            out = batch['output']

            if condition_name not in self.problem.conditions:
                raise RuntimeError('Something wrong happened.')

            # for data driven mode
            if not hasattr(condition, 'output_points'):
                raise NotImplementedError('GAROM works only in data-driven mode.')

            # get data
            snapshots = out[condition_idx == condition_id]
            parameters = pts[condition_idx == condition_id]

            d_loss_real, d_loss_fake, d_loss = self._train_discriminator(
                parameters, snapshots)

            r_loss, g_loss = self._train_generator(parameters, snapshots)
            
            diff = self._update_weights(d_loss_real, d_loss_fake)

            # logging
            self.log('mean_loss', float(r_loss), prog_bar=True, logger=True, on_epoch=True, on_step=False)
            self.log('d_loss', float(d_loss), prog_bar=True, logger=True, on_epoch=True, on_step=False)
            self.log('g_loss', float(g_loss), prog_bar=True, logger=True, on_epoch=True, on_step=False)
            self.log('stability_metric', float(d_loss_real + torch.abs(diff)), prog_bar=True, logger=True, on_epoch=True, on_step=False)

        return

    @property
    def generator(self):
        return self.models[0]

    @property
    def discriminator(self):
        return self.models[1]

    @property
    def optimizer_generator(self):
        return self.optimizers[0]

    @property
    def optimizer_discriminator(self):
        return self.optimizers[1]

    @property
    def scheduler_generator(self):
        return self._schedulers[0]

    @property
    def scheduler_discriminator(self):
        return self._schedulers[1]



"""Module for DeepONet model"""
import torch
import torch.nn as nn
from ..utils import check_consistency, is_function
from functools import partial


class MIONet(torch.nn.Module):
    """
    The PINA implementation of MIONet network.

    MIONet is a general architecture for learning Operators defined
    on the tensor product of Banach spaces. Unlike traditional machine
    learning methods MIONet is designed to map entire functions to other functions.
    It can be trained both with Physics Informed or Supervised learning strategies.

    .. seealso::

        **Original reference**: Jin, Pengzhan, Shuai Meng, and Lu Lu.
        *MIONet: Learning multiple-input operators via tensor product.*
        SIAM Journal on Scientific Computing 44.6 (2022): A3490-A351
        DOI: `10.1137/22M1477751
        <https://doi.org/10.1137/22M1477751>`_

    """

    def __init__(self,
                 networks,
                 aggregator="*",
                 reduction="+",
                 scale=True,
                 translation=True):
        """
        :param dict networks: The neural networks to use as
            models. The ``dict`` takes as key a neural network, and
            as value the list of indeces to extract from the input variable
            in the forward pass of the neural network. If a list of ``int`` is passed,
            the corresponding columns of the inner most entries are extracted.
            If a list of ``str`` is passed the variables of the corresponding :py:obj:`pina.label_tensor.LabelTensor`
            are extracted. The ``torch.nn.Module`` model has to take as input a
            :py:obj:`pina.label_tensor.LabelTensor` or :class:`torch.Tensor`.
            Default implementation consist of different branch nets and one trunk nets.
        :param str or Callable aggregator: Aggregator to be used to aggregate
            partial results from the modules in `nets`. Partial results are
            aggregated component-wise. Available aggregators include
            sum: ``+``, product: ``*``, mean: ``mean``, min: ``min``, max: ``max``.
        :param str or Callable reduction: Reduction to be used to reduce
            the aggregated result of the modules in `nets` to the desired output
            dimension. Available reductions include
            sum: ``+``, product: ``*``, mean: ``mean``, min: ``min``, max: ``max``.
        :param bool or Callable scale: Scaling the final output before returning the
            forward pass, default ``True``.
        :param bool or Callable translation: Translating the final output before
            returning the forward pass, default ``True``.

        .. warning::
            In the forward pass we do not check if the input is instance of
            :py:obj:`pina.label_tensor.LabelTensor` or :class:`torch.Tensor`. A general rule is
            that for a :py:obj:`pina.label_tensor.LabelTensor` input both list of integers and
            list of strings can be passed for ``input_indeces_branch_net``
            and ``input_indeces_trunk_net``. Differently, for a :class:`torch.Tensor`
            only a list of integers can be passed for ``input_indeces_branch_net``
            and ``input_indeces_trunk_net``.

        :Example:
            >>> branch_net1 = FeedForward(input_dimensons=1, output_dimensions=10)
            >>> branch_net2 = FeedForward(input_dimensons=2, output_dimensions=10)
            >>> trunk_net = FeedForward(input_dimensons=1, output_dimensions=10)
            >>> networks = {branch_net1 : ['x'],
                            branch_net2 : ['x', 'y'],
            ...             trunk_net : ['z']}
            >>> model = MIONet(networks=networks,
            ...                reduction='+',
            ...                aggregator='*')
            >>> model
            MIONet(
            (models): ModuleList(
                (0): FeedForward(
                (model): Sequential(
                    (0): Linear(in_features=1, out_features=20, bias=True)
                    (1): Tanh()
                    (2): Linear(in_features=20, out_features=20, bias=True)
                    (3): Tanh()
                    (4): Linear(in_features=20, out_features=10, bias=True)
                )
                )
                (1): FeedForward(
                (model): Sequential(
                    (0): Linear(in_features=2, out_features=20, bias=True)
                    (1): Tanh()
                    (2): Linear(in_features=20, out_features=20, bias=True)
                    (3): Tanh()
                    (4): Linear(in_features=20, out_features=10, bias=True)
                )
                )
                (2): FeedForward(
                (model): Sequential(
                    (0): Linear(in_features=1, out_features=20, bias=True)
                    (1): Tanh()
                    (2): Linear(in_features=20, out_features=20, bias=True)
                    (3): Tanh()
                    (4): Linear(in_features=20, out_features=10, bias=True)
                )
                )
            )
            )
        """
        super().__init__()

        # check type consistency
        check_consistency(networks, dict)
        check_consistency(scale, bool)
        check_consistency(translation, bool)

        # check trunk branch nets consistency
        shapes = []
        for key, value in networks.items():
            check_consistency(value, (str, int))
            check_consistency(key, torch.nn.Module)
            input_ = torch.rand(10, len(value))
            shapes.append(key(input_).shape[-1])

        if not all(map(lambda x: x == shapes[0], shapes)):
            raise ValueError('The passed networks have not the same '
                             'output dimension.')

        # assign trunk and branch net with their input indeces
        self.models = torch.nn.ModuleList(networks.keys())
        self._indeces = networks.values()

        # initializie aggregation
        self._init_aggregator(aggregator=aggregator)
        self._init_reduction(reduction=reduction)

        # scale and translation
        self._scale = torch.nn.Parameter(torch.tensor(
            [1.0])) if scale else torch.tensor([1.0])
        self._trasl = torch.nn.Parameter(torch.tensor(
            [1.0])) if translation else torch.tensor([1.0])

    @staticmethod
    def _symbol_functions(**kwargs):
        """
        Return a dictionary of functions that can be used as aggregators or
        reductions.
        """
        return {
            "+": partial(torch.sum, **kwargs),
            "*": partial(torch.prod, **kwargs),
            "mean": partial(torch.mean, **kwargs),
            "min": lambda x: torch.min(x, **kwargs).values,
            "max": lambda x: torch.max(x, **kwargs).values,
        }

    def _init_aggregator(self, aggregator):
        aggregator_funcs = DeepONet._symbol_functions(dim=2)
        if aggregator in aggregator_funcs:
            aggregator_func = aggregator_funcs[aggregator]
        elif isinstance(aggregator, nn.Module) or is_function(aggregator):
            aggregator_func = aggregator
        else:
            raise ValueError(f"Unsupported aggregation: {str(aggregator)}")

        self._aggregator = aggregator_func

    def _init_reduction(self, reduction):
        reduction_funcs = DeepONet._symbol_functions(dim=-1)
        if reduction in reduction_funcs:
            reduction_func = reduction_funcs[reduction]
        elif isinstance(reduction, nn.Module) or is_function(reduction):
            reduction_func = reduction
        else:
            raise ValueError(f"Unsupported reduction: {reduction}")

        self._reduction = reduction_func

    def _get_vars(self, x, indeces):
        if isinstance(indeces[0], str):
            try:
                return x.extract(indeces)
            except AttributeError:
                raise RuntimeError(
                    'Not possible to extract input variables from tensor.'
                    ' Ensure that the passed tensor is a LabelTensor or'
                    ' pass list of integers to extract variables. For'
                    ' more information refer to warning in the documentation.')
        elif isinstance(indeces[0], int):
            return x[..., indeces]
        else:
            raise RuntimeError(
                'Not able to extract right indeces for tensor.'
                ' For more information refer to warning in the documentation.')

    def forward(self, x):
        """
        Defines the computation performed at every call.

        :param LabelTensor or torch.Tensor x: The input tensor for the forward call.
        :return: The output computed by the DeepONet model.
        :rtype: LabelTensor or torch.Tensor 
        """

        # forward pass
        output_ = [
            model(self._get_vars(x, indeces))
            for model, indeces in zip(self.models, self._indeces)
        ]

        # aggregation
        aggregated = self._aggregator(torch.dstack(output_))

        # reduce
        output_ = self._reduction(aggregated).reshape(-1, 1)

        # scale and translate
        output_ *= self._scale
        output_ += self._trasl

        return output_

    @property
    def aggregator(self):
        """
        The aggregator function.
        """
        return self._aggregator

    @property
    def reduction(self):
        """
        The translation factor.
        """
        return self._reduction

    @property
    def scale(self):
        """
        The scale factor.
        """
        return self._scale

    @property
    def translation(self):
        """
        The translation factor for MIONet.
        """
        return self._trasl

    @property
    def indeces_variables_extracted(self):
        """
        The input indeces for each model in form of list.
        """
        return self._indeces

    @property
    def model(self):
        """
        The models in form of list.
        """
        return self._indeces


class DeepONet(MIONet):
    """
    The PINA implementation of DeepONet network.

    DeepONet is a general architecture for learning Operators. Unlike
    traditional machine learning methods DeepONet is designed to map
    entire functions to other functions. It can be trained both with 
    Physics Informed or Supervised learning strategies.

    .. seealso::

        **Original reference**: Lu, L., Jin, P., Pang, G. et al. *Learning
        nonlinear operators via DeepONet based on the universal approximation
        theorem of operators*. Nat Mach Intell 3, 218–229 (2021).
        DOI: `10.1038/s42256-021-00302-5
        <https://doi.org/10.1038/s42256-021-00302-5>`_

    """

    def __init__(self,
                 branch_net,
                 trunk_net,
                 input_indeces_branch_net,
                 input_indeces_trunk_net,
                 aggregator="*",
                 reduction="+",
                 scale=True,
                 translation=True):
        """
        :param torch.nn.Module branch_net: The neural network to use as branch
            model. It has to take as input a :py:obj:`pina.label_tensor.LabelTensor`
            or :class:`torch.Tensor`. The number of dimensions of the output has
            to be the same of the ``trunk_net``.
        :param torch.nn.Module trunk_net: The neural network to use as trunk
            model. It has to take as input a :py:obj:`pina.label_tensor.LabelTensor`
            or :class:`torch.Tensor`. The number of dimensions of the output
            has to be the same of the ``branch_net``.
        :param list(int) or list(str) input_indeces_branch_net: List of indeces
            to extract from the input variable in the forward pass for the
            branch net. If a list of ``int`` is passed, the corresponding columns
            of the inner most entries are extracted. If a list of ``str`` is passed
            the variables of the corresponding :py:obj:`pina.label_tensor.LabelTensor` are extracted.
        :param list(int) or list(str) input_indeces_trunk_net: List of indeces
            to extract from the input variable in the forward pass for the
            trunk net. If a list of ``int`` is passed, the corresponding columns
            of the inner most entries are extracted. If a list of ``str`` is passed
            the variables of the corresponding :py:obj:`pina.label_tensor.LabelTensor` are extracted.
        :param str or Callable aggregator: Aggregator to be used to aggregate
            partial results from the modules in `nets`. Partial results are
            aggregated component-wise. Available aggregators include
            sum: ``+``, product: ``*``, mean: ``mean``, min: ``min``, max: ``max``.
        :param str or Callable reduction: Reduction to be used to reduce
            the aggregated result of the modules in `nets` to the desired output
            dimension. Available reductions include
            sum: ``+``, product: ``*``, mean: ``mean``, min: ``min``, max: ``max``.
        :param bool or Callable scale: Scaling the final output before returning the
            forward pass, default True.
        :param bool or Callable translation: Translating the final output before
            returning the forward pass, default True.

        .. warning::
            In the forward pass we do not check if the input is instance of
            :py:obj:`pina.label_tensor.LabelTensor` or :class:`torch.Tensor`. A general rule is
            that for a :py:obj:`pina.label_tensor.LabelTensor` input both list of integers and
            list of strings can be passed for ``input_indeces_branch_net``
            and ``input_indeces_trunk_net``. Differently, for a :class:`torch.Tensor`
            only a list of integers can be passed for ``input_indeces_branch_net``
            and ``input_indeces_trunk_net``.

        :Example:
            >>> branch_net = FeedForward(input_dimensons=1, output_dimensions=10)
            >>> trunk_net = FeedForward(input_dimensons=1, output_dimensions=10)
            >>> model = DeepONet(branch_net=branch_net,
            ...                  trunk_net=trunk_net,
            ...                  input_indeces_branch_net=['x'],
            ...                  input_indeces_trunk_net=['t'],
            ...                  reduction='+',
            ...                  aggregator='*')
            >>> model
            DeepONet(
            (trunk_net): FeedForward(
                (model): Sequential(
                (0): Linear(in_features=1, out_features=20, bias=True)
                (1): Tanh()
                (2): Linear(in_features=20, out_features=20, bias=True)
                (3): Tanh()
                (4): Linear(in_features=20, out_features=10, bias=True)
                )
            )
            (branch_net): FeedForward(
                (model): Sequential(
                (0): Linear(in_features=1, out_features=20, bias=True)
                (1): Tanh()
                (2): Linear(in_features=20, out_features=20, bias=True)
                (3): Tanh()
                (4): Linear(in_features=20, out_features=10, bias=True)
                )
            )
            )
        """
        networks = {
            branch_net: input_indeces_branch_net,
            trunk_net: input_indeces_trunk_net
        }
        super().__init__(networks=networks,
                         aggregator=aggregator,
                         reduction=reduction,
                         scale=scale,
                         translation=translation)
        

    def forward(self, x):
        """
        Defines the computation performed at every call.

        :param LabelTensor or torch.Tensor x: The input tensor for the forward call.
        :return: The output computed by the DeepONet model.
        :rtype: LabelTensor or torch.Tensor 
        """
        return super().forward(x)


    @property
    def branch_net(self):
        """
        The branch net for DeepONet.
        """
        return self.models[0]

    @property
    def trunk_net(self):
        """
        The trunk net for DeepONet.
        """
        return self.models[1]



""" Trainer module. """

import pytorch_lightning
from .utils import check_consistency
from .dataset import SamplePointDataset, SamplePointLoader, DataPointDataset
from .solvers.solver import SolverInterface

class Trainer(pytorch_lightning.Trainer):

    def __init__(self, solver, batch_size=None, **kwargs):
        """
        PINA Trainer class for costumizing every aspect of training via flags.

        :param solver: A pina:class:`SolverInterface` solver for the differential problem.
        :type solver: SolverInterface
        :param batch_size: How many samples per batch to load. If ``batch_size=None`` all
            samples are loaded and data are not batched, defaults to None.
        :type batch_size: int | None

        :Keyword Arguments:
            The additional keyword arguments specify the training setup
            and can be choosen from the `pytorch-lightning
            Trainer API <https://lightning.ai/docs/pytorch/stable/common/trainer.html#trainer-class-api>`_
        """

        super().__init__(**kwargs)

        # check inheritance consistency for solver and batch size
        check_consistency(solver, SolverInterface)
        if batch_size is not None:
            check_consistency(batch_size, int)
        
        self._model = solver
        self.batch_size = batch_size

        # create dataloader
        if solver.problem.have_sampled_points is False:
            raise RuntimeError(f'Input points in {solver.problem.not_sampled_points} '
                               'training are None. Please '
                               'sample points in your problem by calling '
                               'discretise_domain function before train '
                               'in the provided locations.')
        
        self._create_or_update_loader()

    def _create_or_update_loader(self):
        """
        This method is used here because is resampling is needed
        during training, there is no need to define to touch the
        trainer dataloader, just call the method.
        """
        devices = self._accelerator_connector._parallel_devices

        if len(devices) > 1:
            raise RuntimeError('Parallel training is not supported yet.')

        device = devices[0]
        dataset_phys = SamplePointDataset(self._model.problem, device)
        dataset_data = DataPointDataset(self._model.problem, device)
        self._loader = SamplePointLoader(
            dataset_phys, dataset_data, batch_size=self.batch_size,
            shuffle=True)

    def train(self, **kwargs):
        """
        Train the solver method.
        """
        return super().fit(self._model, train_dataloaders=self._loader, **kwargs)
    
    @property
    def solver(self):
        """
        Returning trainer solver.
        """
        return self._model



'''PINA Callbacks Implementations'''

from pytorch_lightning.callbacks import Callback
import torch
from ..utils import check_consistency


class SwitchOptimizer(Callback):

    def __init__(self, new_optimizers, new_optimizers_kwargs, epoch_switch):
        """
        PINA Implementation of a Lightning Callback to switch optimizer during training.

        This callback allows for switching between different optimizers during training, enabling
        the exploration of multiple optimization strategies without the need to stop training.

        :param new_optimizers: The model optimizers to switch to. Can be a single 
                            :class:`torch.optim.Optimizer` or a list of them for multiple model solvers.
        :type new_optimizers: torch.optim.Optimizer | list
        :param new_optimizers_kwargs: The keyword arguments for the new optimizers. Can be a single dictionary
                                    or a list of dictionaries corresponding to each optimizer.
        :type new_optimizers_kwargs: dict | list
        :param epoch_switch: The epoch at which to switch to the new optimizer.
        :type epoch_switch: int

        :raises ValueError: If `epoch_switch` is less than 1 or if there is a mismatch in the number of 
                            optimizers and their corresponding keyword argument dictionaries.

        Example:
            >>> switch_callback = SwitchOptimizer(new_optimizers=[optimizer1, optimizer2],
            >>>                                  new_optimizers_kwargs=[{'lr': 0.001}, {'lr': 0.01}],
            >>>                                  epoch_switch=10)
        """
        super().__init__()

        # check type consistency
        check_consistency(new_optimizers, torch.optim.Optimizer, subclass=True)
        check_consistency(new_optimizers_kwargs, dict)
        check_consistency(epoch_switch, int)

        if epoch_switch < 1:
            raise ValueError('epoch_switch must be greater than one.')

        if not isinstance(new_optimizers, list):
            new_optimizers = [new_optimizers]
            new_optimizers_kwargs = [new_optimizers_kwargs]
        len_optimizer = len(new_optimizers)
        len_optimizer_kwargs = len(new_optimizers_kwargs)

        if len_optimizer_kwargs != len_optimizer:
            raise ValueError('You must define one dictionary of keyword'
                             ' arguments for each optimizers.'
                             f' Got {len_optimizer} optimizers, and'
                             f' {len_optimizer_kwargs} dicitionaries')

        # save new optimizers
        self._new_optimizers = new_optimizers
        self._new_optimizers_kwargs = new_optimizers_kwargs
        self._epoch_switch = epoch_switch

    def on_train_epoch_start(self, trainer, __):
        """
        Callback function to switch optimizer at the start of each training epoch.

        :param trainer: The trainer object managing the training process.
        :type trainer: pytorch_lightning.Trainer
        :param _: Placeholder argument (not used).

        :return: None
        :rtype: None
        """
        if trainer.current_epoch == self._epoch_switch:
            optims = []
            for idx, (optim, optim_kwargs) in enumerate(
                    zip(self._new_optimizers, self._new_optimizers_kwargs)):
                optims.append(
                    optim(trainer._model.models[idx].parameters(),
                          **optim_kwargs))

            trainer.optimizers = optims



""" Module for OperationInterface class. """

from .location import Location
from ..utils import check_consistency
from abc import ABCMeta, abstractmethod


class OperationInterface(Location, metaclass=ABCMeta):

    def __init__(self, geometries):
        """
        Abstract set operation class. Any geometry operation entity must inherit from this class.

        :param list geometries: A list of geometries from ``pina.geometry``
            such as ``EllipsoidDomain`` or ``CartesianDomain``.
        """
        # check consistency geometries
        check_consistency(geometries, Location)

        # check we are passing always different
        # geometries with the same labels.
        self._check_dimensions(geometries)

        # assign geometries
        self._geometries = geometries

    @property
    def geometries(self):
        """ 
        The geometries to perform set operation.
        """
        return self._geometries

    @property
    def variables(self):
        """
        Spatial variables of the domain.

        :return: All the variables defined in ``__init__`` in order.
        :rtype: list[str]
        """
        return self.geometries[0].variables
    
    @ abstractmethod
    def is_inside(self, point, check_border=False):
        """
        Check if a point is inside the resulting domain after
        a set operation is applied.

        :param point: Point to be checked.
        :type point: torch.Tensor   
        :param bool check_border: If ``True``, the border is considered inside.
        :return: ``True`` if the point is inside the Intersection domain, ``False`` otherwise.
        :rtype: bool
        """
        pass

    def _check_dimensions(self, geometries):
        """Check if the dimensions of the geometries are consistent.

        :param geometries: Geometries to be checked.
        :type geometries: list[Location]
        """
        for geometry in geometries:
            if geometry.variables != geometries[0].variables:
                raise NotImplementedError(
                    f'The geometries need to have same dimensions and labels.')



""" Module for EquationInterface class """
from abc import ABCMeta, abstractmethod


class EquationInterface(metaclass=ABCMeta):
    """
    The abstract `AbstractProblem` class. All the class defining a PINA Problem
    should be inheritied from this class.

    In the definition of a PINA problem, the fundamental elements are:
    the output variables, the condition(s), and the domain(s) where the
    conditions are applied.
    """

    @abstractmethod
    def residual(self, input_, output_, params_):
        """
        Residual computation of the equation.

        :param LabelTensor input_: Input points to evaluate the equation.
        :param LabelTensor output_: Output vectors given by my model (e.g., a ``FeedForward`` model).
        :param dict params_: Dictionary of unknown parameters, eventually
            related to an ``InverseProblem``.
        :return: The residual evaluation of the specified equation.
        :rtype: LabelTensor
        """
        pass



import torch
import torch.nn as nn
from ...utils import check_consistency
import warnings


######## 1D Spectral Convolution ###########
class SpectralConvBlock1D(nn.Module):
    """
    PINA implementation of Spectral Convolution Block for one
    dimensional tensors.
    """

    def __init__(self, input_numb_fields, output_numb_fields, n_modes):
        """
        The module computes the spectral convolution of the input with a linear kernel in the
        fourier space, and then it maps the input back to the physical
        space.

        The block expects an input of size ``[batch, input_numb_fields, N]``
        and returns an output of size ``[batch, output_numb_fields, N]``.

        :param int input_numb_fields: The number of channels for the input.
        :param int output_numb_fields: The number of channels for the output.
        :param int n_modes: Number of modes to select, it must be at most equal
            to the ``floor(N/2)+1``.
        """
        super().__init__()

        # check type consistency
        check_consistency(input_numb_fields, int)
        check_consistency(output_numb_fields, int)

        # assign variables
        self._modes = n_modes
        self._input_channels = input_numb_fields
        self._output_channels = output_numb_fields

        # scaling factor
        scale = (1. / (self._input_channels * self._output_channels))
        self._weights = nn.Parameter(scale * torch.rand(self._input_channels,
                                                        self._output_channels,
                                                        self._modes,
                                                        dtype=torch.cfloat))

    def _compute_mult1d(self, input, weights):
        """
        Compute the matrix multiplication of the input
        with the linear kernel weights.

        :param input: The input tensor, expect of size 
            ``[batch, input_numb_fields, x]``.
        :type input: torch.Tensor
        :param weights: The kernel weights, expect of
            size ``[input_numb_fields, output_numb_fields, x]``.
        :type weights: torch.Tensor
        :return: The matrix multiplication of the input
            with the linear kernel weights.
        :rtype: torch.Tensor
        """
        return torch.einsum("bix,iox->box", input, weights)

    def forward(self, x):
        """
        Forward computation for Spectral Convolution.

        :param x: The input tensor, expect of size 
            ``[batch, input_numb_fields, x]``.
        :type x: torch.Tensor
        :return: The output tensor obtained from the
            spectral convolution of size ``[batch, output_numb_fields, x]``.
        :rtype: torch.Tensor
        """
        batch_size = x.shape[0]

        # Compute Fourier transform of the input
        x_ft = torch.fft.rfft(x)

        # Multiply relevant Fourier modes
        out_ft = torch.zeros(batch_size,
                             self._output_channels,
                             x.size(-1) // 2 + 1,
                             device=x.device,
                             dtype=torch.cfloat)
        out_ft[:, :, :self._modes] = self._compute_mult1d(
            x_ft[:, :, :self._modes], self._weights)

        # Return to physical space
        return torch.fft.irfft(out_ft, n=x.size(-1))


######## 2D Spectral Convolution ###########
class SpectralConvBlock2D(nn.Module):
    """
    PINA implementation of spectral convolution block for two
    dimensional tensors.
    """

    def __init__(self, input_numb_fields, output_numb_fields, n_modes):
        """
        The module computes the spectral convolution of the input with a linear kernel in the
        fourier space, and then it maps the input back to the physical
        space.

        The block expects an input of size ``[batch, input_numb_fields, Nx, Ny]``
        and returns an output of size ``[batch, output_numb_fields, Nx, Ny]``.

        :param int input_numb_fields: The number of channels for the input.
        :param int output_numb_fields: The number of channels for the output.
        :param list | tuple n_modes: Number of modes to select for each dimension.
            It must be at most equal to the ``floor(Nx/2)+1`` and ``floor(Ny/2)+1``.
        """
        super().__init__()

        # check type consistency
        check_consistency(input_numb_fields, int)
        check_consistency(output_numb_fields, int)
        check_consistency(n_modes, int)
        if isinstance(n_modes, (tuple, list)):
            if len(n_modes) != 2:
                raise ValueError(
                    'Expected n_modes to be a list or tuple of len two, '
                    'with each entry corresponding to the number of modes '
                    'for each dimension ')
        elif isinstance(n_modes, int):
            n_modes = [n_modes] * 2
        else:
            raise ValueError(
                'Expected n_modes to be a list or tuple of len two, '
                'with each entry corresponding to the number of modes '
                'for each dimension; or an int value representing the '
                'number of modes for all dimensions')

        # assign variables
        self._modes = n_modes
        self._input_channels = input_numb_fields
        self._output_channels = output_numb_fields

        # scaling factor
        scale = (1. / (self._input_channels * self._output_channels))
        self._weights1 = nn.Parameter(scale * torch.rand(self._input_channels,
                                                         self._output_channels,
                                                         self._modes[0],
                                                         self._modes[1],
                                                         dtype=torch.cfloat))
        self._weights2 = nn.Parameter(scale * torch.rand(self._input_channels,
                                                         self._output_channels,
                                                         self._modes[0],
                                                         self._modes[1],
                                                         dtype=torch.cfloat))

    def _compute_mult2d(self, input, weights):
        """
        Compute the matrix multiplication of the input
        with the linear kernel weights.

        :param input: The input tensor, expect of size 
            ``[batch, input_numb_fields, x, y]``.
        :type input: torch.Tensor
        :param weights: The kernel weights, expect of
            size ``[input_numb_fields, output_numb_fields, x, y]``.
        :type weights: torch.Tensor
        :return: The matrix multiplication of the input
            with the linear kernel weights.
        :rtype: torch.Tensor
        """
        return torch.einsum("bixy,ioxy->boxy", input, weights)

    def forward(self, x):
        """
        Forward computation for Spectral Convolution.

        :param x: The input tensor, expect of size 
            ``[batch, input_numb_fields, x, y]``.
        :type x: torch.Tensor
        :return: The output tensor obtained from the
            spectral convolution of size ``[batch, output_numb_fields, x, y]``.
        :rtype: torch.Tensor
        """

        batch_size = x.shape[0]

        # Compute Fourier transform of the input
        x_ft = torch.fft.rfft2(x)

        # Multiply relevant Fourier modes
        out_ft = torch.zeros(batch_size,
                             self._output_channels,
                             x.size(-2),
                             x.size(-1) // 2 + 1,
                             device=x.device,
                             dtype=torch.cfloat)
        out_ft[:, :, :self._modes[0], :self._modes[1]] = self._compute_mult2d(
            x_ft[:, :, :self._modes[0], :self._modes[1]], self._weights1)
        out_ft[:, :, -self._modes[0]:, :self._modes[1]:] = self._compute_mult2d(
            x_ft[:, :, -self._modes[0]:, :self._modes[1]], self._weights2)

        # Return to physical space
        return torch.fft.irfft2(out_ft, s=(x.size(-2), x.size(-1)))


######## 3D Spectral Convolution ###########
class SpectralConvBlock3D(nn.Module):
    """
    PINA implementation of spectral convolution block for three
    dimensional tensors.
    """

    def __init__(self, input_numb_fields, output_numb_fields, n_modes):
        """
        The module computes the spectral convolution of the input with a linear kernel in the
        fourier space, and then it maps the input back to the physical
        space.

        The block expects an input of size ``[batch, input_numb_fields, Nx, Ny, Nz]``
        and returns an output of size ``[batch, output_numb_fields, Nx, Ny, Nz]``.

        :param int input_numb_fields: The number of channels for the input.
        :param int output_numb_fields: The number of channels for the output.
        :param list | tuple n_modes: Number of modes to select for each dimension.
            It must be at most equal to the ``floor(Nx/2)+1``, ``floor(Ny/2)+1``
            and ``floor(Nz/2)+1``.
        """
        super().__init__()

        # check type consistency
        check_consistency(input_numb_fields, int)
        check_consistency(output_numb_fields, int)
        check_consistency(n_modes, int)
        if isinstance(n_modes, (tuple, list)):
            if len(n_modes) != 3:
                raise ValueError(
                    'Expected n_modes to be a list or tuple of len three, '
                    'with each entry corresponding to the number of modes '
                    'for each dimension ')
        elif isinstance(n_modes, int):
            n_modes = [n_modes] * 3
        else:
            raise ValueError(
                'Expected n_modes to be a list or tuple of len three, '
                'with each entry corresponding to the number of modes '
                'for each dimension; or an int value representing the '
                'number of modes for all dimensions')

        # assign variables
        self._modes = n_modes
        self._input_channels = input_numb_fields
        self._output_channels = output_numb_fields

        # scaling factor
        scale = (1. / (self._input_channels * self._output_channels))
        self._weights1 = nn.Parameter(scale * torch.rand(self._input_channels,
                                                         self._output_channels,
                                                         self._modes[0],
                                                         self._modes[1],
                                                         self._modes[2],
                                                         dtype=torch.cfloat))
        self._weights2 = nn.Parameter(scale * torch.rand(self._input_channels,
                                                         self._output_channels,
                                                         self._modes[0],
                                                         self._modes[1],
                                                         self._modes[2],
                                                         dtype=torch.cfloat))
        self._weights3 = nn.Parameter(scale * torch.rand(self._input_channels,
                                                         self._output_channels,
                                                         self._modes[0],
                                                         self._modes[1],
                                                         self._modes[2],
                                                         dtype=torch.cfloat))
        self._weights4 = nn.Parameter(scale * torch.rand(self._input_channels,
                                                         self._output_channels,
                                                         self._modes[0],
                                                         self._modes[1],
                                                         self._modes[2],
                                                         dtype=torch.cfloat))

    def _compute_mult3d(self, input, weights):
        """
        Compute the matrix multiplication of the input
        with the linear kernel weights.

        :param input: The input tensor, expect of size 
            ``[batch, input_numb_fields, x, y, z]``.
        :type input: torch.Tensor
        :param weights: The kernel weights, expect of
            size ``[input_numb_fields, output_numb_fields, x, y, z]``.
        :type weights: torch.Tensor
        :return: The matrix multiplication of the input
            with the linear kernel weights.
        :rtype: torch.Tensor
        """
        return torch.einsum("bixyz,ioxyz->boxyz", input, weights)

    def forward(self, x):
        """
        Forward computation for Spectral Convolution.

        :param x: The input tensor, expect of size 
            ``[batch, input_numb_fields, x, y, z]``.
        :type x: torch.Tensor
        :return: The output tensor obtained from the
            spectral convolution of size ``[batch, output_numb_fields, x, y, z]``.
        :rtype: torch.Tensor
        """

        batch_size = x.shape[0]

        # Compute Fourier transform of the input
        x_ft = torch.fft.rfftn(x, dim=[-3, -2, -1])

        # Multiply relevant Fourier modes
        out_ft = torch.zeros(batch_size,
                             self._output_channels,
                             x.size(-3),
                             x.size(-2),
                             x.size(-1) // 2 + 1,
                             device=x.device,
                             dtype=torch.cfloat)

        slice0 = (
            slice(None),
            slice(None),
            slice(self._modes[0]),
            slice(self._modes[1]),
            slice(self._modes[2]),
        )
        out_ft[slice0] = self._compute_mult3d(x_ft[slice0], self._weights1)

        slice1 = (
            slice(None),
            slice(None),
            slice(self._modes[0]),
            slice(-self._modes[1], None),
            slice(self._modes[2]),
        )
        out_ft[slice1] = self._compute_mult3d(x_ft[slice1], self._weights2)

        slice2 = (
            slice(None),
            slice(None),
            slice(-self._modes[0], None),
            slice(self._modes[1]),
            slice(self._modes[2]),
        )
        out_ft[slice2] = self._compute_mult3d(x_ft[slice2], self._weights3)

        slice3 = (
            slice(None),
            slice(None),
            slice(-self._modes[0], None),
            slice(-self._modes[1], None),
            slice(self._modes[2]),
        )
        out_ft[slice3] = self._compute_mult3d(x_ft[slice3], self._weights4)

        # Return to physical space
        return torch.fft.irfftn(out_ft, s=(x.size(-3), x.size(-2), x.size(-1)))



import torch
import pytest

from pina import LabelTensor

data = torch.rand((20, 3))
labels = ['a', 'b', 'c']


def test_constructor():
    LabelTensor(data, labels)


def test_wrong_constructor():
    with pytest.raises(ValueError):
        LabelTensor(data, ['a', 'b'])


def test_labels():
    tensor = LabelTensor(data, labels)
    assert isinstance(tensor, torch.Tensor)
    assert tensor.labels == labels
    with pytest.raises(ValueError):
        tensor.labels = labels[:-1]


def test_extract():
    label_to_extract = ['a', 'c']
    tensor = LabelTensor(data, labels)
    new = tensor.extract(label_to_extract)
    assert new.labels == label_to_extract
    assert new.shape[1] == len(label_to_extract)
    assert torch.all(torch.isclose(data[:, 0::2], new))


def test_extract_onelabel():
    label_to_extract = ['a']
    tensor = LabelTensor(data, labels)
    new = tensor.extract(label_to_extract)
    assert new.ndim == 2
    assert new.labels == label_to_extract
    assert new.shape[1] == len(label_to_extract)
    assert torch.all(torch.isclose(data[:, 0].reshape(-1, 1), new))


def test_wrong_extract():
    label_to_extract = ['a', 'cc']
    tensor = LabelTensor(data, labels)
    with pytest.raises(ValueError):
        tensor.extract(label_to_extract)


def test_extract_order():
    label_to_extract = ['c', 'a']
    tensor = LabelTensor(data, labels)
    new = tensor.extract(label_to_extract)
    expected = torch.cat(
        (data[:, 2].reshape(-1, 1), data[:, 0].reshape(-1, 1)),
        dim=1)
    assert new.labels == label_to_extract
    assert new.shape[1] == len(label_to_extract)
    assert torch.all(torch.isclose(expected, new))


def test_merge():
    tensor = LabelTensor(data, labels)
    tensor_a = tensor.extract('a')
    tensor_b = tensor.extract('b')
    tensor_c = tensor.extract('c')

    tensor_bc = tensor_b.append(tensor_c)
    assert torch.allclose(tensor_bc, tensor.extract(['b', 'c']))


def test_merge2():
    tensor = LabelTensor(data, labels)
    tensor_b = tensor.extract('b')
    tensor_c = tensor.extract('c')

    tensor_bc = tensor_b.append(tensor_c)
    assert torch.allclose(tensor_bc, tensor.extract(['b', 'c']))


def test_getitem():
    tensor = LabelTensor(data, labels)
    tensor_view = tensor['a']

    assert tensor_view.labels == ['a']
    assert torch.allclose(tensor_view.flatten(), data[:, 0])

    tensor_view = tensor['a', 'c']

    assert tensor_view.labels == ['a', 'c']
    assert torch.allclose(tensor_view, data[:, 0::2])

def test_getitem2():
    tensor = LabelTensor(data, labels)
    tensor_view = tensor[:5]
    assert tensor_view.labels == labels
    assert torch.allclose(tensor_view, data[:5])

    idx = torch.randperm(tensor.shape[0])
    tensor_view = tensor[idx]
    assert tensor_view.labels == labels


def test_slice():
    tensor = LabelTensor(data, labels)
    tensor_view = tensor[:5, :2]
    assert tensor_view.labels == labels[:2]
    assert torch.allclose(tensor_view, data[:5, :2])

    tensor_view2 = tensor[3]
    assert tensor_view2.labels == labels
    assert torch.allclose(tensor_view2, data[3])

    tensor_view3 = tensor[:, 2]
    assert tensor_view3.labels == labels[2]
    assert torch.allclose(tensor_view3, data[:, 2].reshape(-1, 1))



import torch

from .location import Location
from ..label_tensor import LabelTensor
from ..utils import torch_lhs, chebyshev_roots


class CartesianDomain(Location):
    """PINA implementation of Hypercube domain."""

    def __init__(self, cartesian_dict):
        """
        :param cartesian_dict: A dictionary with dict-key a string representing
            the input variables for the pinn, and dict-value a list with
            the domain extrema.
        :type cartesian_dict: dict

        :Example:
            >>> spatial_domain = CartesianDomain({'x': [0, 1], 'y': [0, 1]})
        """
        self.fixed_ = {}
        self.range_ = {}

        for k, v in cartesian_dict.items():
            if isinstance(v, (int, float)):
                self.fixed_[k] = v
            elif isinstance(v, (list, tuple)) and len(v) == 2:
                self.range_[k] = v
            else:
                raise TypeError

    @property
    def variables(self):
        """Spatial variables.

        :return: Spatial variables defined in ``__init__()``
        :rtype: list[str]
        """
        return list(self.fixed_.keys()) + list(self.range_.keys())

    def update(self, new_domain):
        """Adding new dimensions on the ``CartesianDomain``

        :param CartesianDomain new_domain: A new ``CartesianDomain`` object to merge

        :Example:
            >>> spatial_domain = CartesianDomain({'x': [0, 1], 'y': [0, 1]})
            >>> spatial_domain.variables
            ['x', 'y']
            >>> spatial_domain_2 = CartesianDomain({'z': [3, 4], 'w': [0, 1]})
            >>> spatial_domain.update(spatial_domain_2)
            >>> spatial_domain.variables
            ['x', 'y', 'z', 'w']
        """
        self.fixed_.update(new_domain.fixed_)
        self.range_.update(new_domain.range_)

    def _sample_range(self, n, mode, bounds):
        """Rescale the samples to the correct bounds

        :param n: Number of points to sample, see Note below
            for reference.
        :type n: int
        :param mode: Mode for sampling, defaults to ``random``.
            Available modes include: random sampling, ``random``;
            latin hypercube sampling, ``latin`` or ``lh``;
            chebyshev sampling, ``chebyshev``; grid sampling ``grid``.
        :type mode: str
        :param bounds: Bounds to rescale the samples.
        :type bounds: torch.Tensor
        :return: Rescaled sample points.
        :rtype: torch.Tensor
        """
        dim = bounds.shape[0]
        if mode in ['chebyshev', 'grid'] and dim != 1:
            raise RuntimeError('Something wrong in Span...')

        if mode == 'random':
            pts = torch.rand(size=(n, dim))
        elif mode == 'chebyshev':
            pts = chebyshev_roots(n).mul(.5).add(.5).reshape(-1, 1)
        elif mode == 'grid':
            pts = torch.linspace(0, 1, n).reshape(-1, 1)
        # elif mode == 'lh' or mode == 'latin':
        elif mode in ['lh', 'latin']:
            pts = torch_lhs(n, dim)

        pts *= bounds[:, 1] - bounds[:, 0]
        pts += bounds[:, 0]

        return pts

    def sample(self, n, mode='random', variables='all'):
        """Sample routine.

        :param n: Number of points to sample, see Note below
            for reference.
        :type n: int
        :param mode: Mode for sampling, defaults to ``random``.
            Available modes include: random sampling, ``random``;
            latin hypercube sampling, ``latin`` or ``lh``;
            chebyshev sampling, ``chebyshev``; grid sampling ``grid``.
        :type mode: str
        :param variables: pinn variable to be sampled, defaults to ``all``.
        :type variables: str | list[str]
        :return: Returns ``LabelTensor`` of n sampled points.
        :rtype: LabelTensor

        .. note::
            The total number of points sampled in case of multiple variables
            is not ``n``, and it depends on the chosen ``mode``. If ``mode`` is
            'grid' or ``chebyshev``, the points are sampled independentely
            across the variables and the results crossed together, i.e. the
            final number of points is ``n`` to the power of the number of
            variables. If 'mode' is 'random', ``lh`` or ``latin``, the variables
            are sampled all together, and the final number of points

        .. warning::
            The extrema values of Span are always sampled only for ``grid`` mode.

        :Example:
            >>> spatial_domain = Span({'x': [0, 1], 'y': [0, 1]})
            >>> spatial_domain.sample(n=4, mode='random')
                tensor([[0.0108, 0.7643],
                        [0.4477, 0.8015],
                        [0.2063, 0.8087],
                        [0.8735, 0.6349]])
            >>> spatial_domain.sample(n=4, mode='grid')
                tensor([[0.0000, 0.0000],
                        [0.3333, 0.0000],
                        [0.6667, 0.0000],
                        [1.0000, 0.0000],
                        [0.0000, 0.3333],
                        [0.3333, 0.3333],
                        [0.6667, 0.3333],
                        [1.0000, 0.3333],
                        [0.0000, 0.6667],
                        [0.3333, 0.6667],
                        [0.6667, 0.6667],
                        [1.0000, 0.6667],
                        [0.0000, 1.0000],
                        [0.3333, 1.0000],
                        [0.6667, 1.0000],
                        [1.0000, 1.0000]])
        """

        def _1d_sampler(n, mode, variables):
            """ Sample independentely the variables and cross the results"""
            tmp = []
            for variable in variables:
                if variable in self.range_.keys():
                    bound = torch.tensor([self.range_[variable]])
                    pts_variable = self._sample_range(n, mode, bound)
                    pts_variable = pts_variable.as_subclass(LabelTensor)
                    pts_variable.labels = [variable]

                    tmp.append(pts_variable)

            result = tmp[0]
            for i in tmp[1:]:
                result = result.append(i, mode='cross')

            for variable in variables:
                if variable in self.fixed_.keys():
                    value = self.fixed_[variable]
                    pts_variable = torch.tensor([[value]
                                                 ]).repeat(result.shape[0], 1)
                    pts_variable = pts_variable.as_subclass(LabelTensor)
                    pts_variable.labels = [variable]

                    result = result.append(pts_variable, mode='std')

            return result

        def _Nd_sampler(n, mode, variables):
            """Sample all the variables together

            :param n: Number of points to sample.
            :type n: int
            :param mode: Mode for sampling, defaults to ``random``.
                Available modes include: random sampling, ``random``;
                latin hypercube sampling, ``latin`` or ``lh``;
                chebyshev sampling, ``chebyshev``; grid sampling ``grid``.
            :type mode: str.
            :param variables: pinn variable to be sampled, defaults to ``all``.
            :type variables: str or list[str].
            :return: Sample points.
            :rtype: list[torch.Tensor]
            """
            pairs = [(k, v) for k, v in self.range_.items() if k in variables]
            keys, values = map(list, zip(*pairs))
            bounds = torch.tensor(values)
            result = self._sample_range(n, mode, bounds)
            result = result.as_subclass(LabelTensor)
            result.labels = keys

            for variable in variables:
                if variable in self.fixed_.keys():
                    value = self.fixed_[variable]
                    pts_variable = torch.tensor([[value]
                                                 ]).repeat(result.shape[0], 1)
                    pts_variable = pts_variable.as_subclass(LabelTensor)
                    pts_variable.labels = [variable]

                    result = result.append(pts_variable, mode='std')
            return result

        def _single_points_sample(n, variables):
            """Sample a single point in one dimension.

            :param n: Number of points to sample.
            :type n: int
            :param variables: Variables to sample from.
            :type variables: list[str]
            :return: Sample points.
            :rtype: list[torch.Tensor]
            """
            tmp = []
            for variable in variables:
                if variable in self.fixed_.keys():
                    value = self.fixed_[variable]
                    pts_variable = torch.tensor([[value]]).repeat(n, 1)
                    pts_variable = pts_variable.as_subclass(LabelTensor)
                    pts_variable.labels = [variable]
                    tmp.append(pts_variable)

            result = tmp[0]
            for i in tmp[1:]:
                result = result.append(i, mode='std')

            return result

        if self.fixed_ and (not self.range_):
            return _single_points_sample(n, variables)

        if variables == 'all':
            variables = list(self.range_.keys()) + list(self.fixed_.keys())

        if mode in ['grid', 'chebyshev']:
            return _1d_sampler(n, mode, variables)
        elif mode in ['random', 'lh', 'latin']:
            return _Nd_sampler(n, mode, variables)
        else:
            raise ValueError(f'mode={mode} is not valid.')

    def is_inside(self, point, check_border=False):
        """Check if a point is inside the ellipsoid.

        :param point: Point to be checked
        :type point: LabelTensor
        :param check_border: Check if the point is also on the frontier
            of the hypercube, default ``False``.
        :type check_border: bool
        :return: Returning ``True`` if the point is inside, ``False`` otherwise.
        :rtype: bool
        """
        is_inside = []

        # check fixed variables
        for variable, value in self.fixed_.items():
            if variable in point.labels:
                is_inside.append(point.extract([variable]) == value)

        # check not fixed variables
        for variable, bound in self.range_.items():
            if variable in point.labels:

                if check_border:
                    check = bound[0] <= point.extract([variable]) <= bound[1]
                else:
                    check = bound[0] < point.extract([variable]) < bound[1]

                is_inside.append(check)

        return all(is_inside)



import torch
import pytest

from pina import LabelTensor
from pina.operators import grad, div, laplacian


def func_vec(x):
    return x**2


def func_scalar(x):
    print('X')
    x_ = x.extract(['x'])
    y_ = x.extract(['y'])
    mu_ = x.extract(['mu'])
    return x_**2 + y_**2 + mu_**3


data = torch.rand((20, 3), requires_grad=True)
inp = LabelTensor(data, ['x', 'y', 'mu'])
labels = ['a', 'b', 'c']
tensor_v = LabelTensor(func_vec(inp), labels)
tensor_s = LabelTensor(func_scalar(inp).reshape(-1, 1), labels[0])


def test_grad_scalar_output():
    grad_tensor_s = grad(tensor_s, inp)
    assert grad_tensor_s.shape == inp.shape
    assert grad_tensor_s.labels == [
        f'd{tensor_s.labels[0]}d{i}' for i in inp.labels
    ]
    grad_tensor_s = grad(tensor_s, inp, d=['x', 'y'])
    assert grad_tensor_s.shape == (inp.shape[0], 2)
    assert grad_tensor_s.labels == [
        f'd{tensor_s.labels[0]}d{i}' for i in ['x', 'y']
    ]


def test_grad_vector_output():
    grad_tensor_v = grad(tensor_v, inp)
    assert grad_tensor_v.shape == (20, 9)
    grad_tensor_v = grad(tensor_v, inp, d=['x', 'mu'])
    assert grad_tensor_v.shape == (inp.shape[0], 6)


def test_div_vector_output():
    grad_tensor_v = div(tensor_v, inp)
    assert grad_tensor_v.shape == (20, 1)
    grad_tensor_v = div(tensor_v, inp, components=['a', 'b'], d=['x', 'mu'])
    assert grad_tensor_v.shape == (inp.shape[0], 1)


def test_laplacian_scalar_output():
    laplace_tensor_v = laplacian(tensor_s, inp, components=['a'], d=['x', 'y'])
    assert laplace_tensor_v.shape == tensor_s.shape


def test_laplacian_vector_output():
    laplace_tensor_v = laplacian(tensor_v, inp)
    assert laplace_tensor_v.shape == tensor_v.shape
    laplace_tensor_v = laplacian(tensor_v,
                                 inp,
                                 components=['a', 'b'],
                                 d=['x', 'y'])
    assert laplace_tensor_v.shape == tensor_v.extract(['a', 'b']).shape



__all__ = [
    'ContinuousConvBlock',
    'ResidualBlock',
    'EnhancedLinear',
    'SpectralConvBlock1D',
    'SpectralConvBlock2D',
    'SpectralConvBlock3D',
    'FourierBlock1D',
    'FourierBlock2D',
    'FourierBlock3D',
]

from .convolution_2d import ContinuousConvBlock
from .residual import ResidualBlock, EnhancedLinear
from .spectral import SpectralConvBlock1D, SpectralConvBlock2D, SpectralConvBlock3D
from .fourier import FourierBlock1D, FourierBlock2D, FourierBlock3D



""" Module for plotting. """

import matplotlib.pyplot as plt
import torch
from pina.callbacks import MetricTracker
from pina import LabelTensor


class Plotter:
    """
    Implementation of a plotter class, for easy visualizations.
    """

    def plot_samples(self, problem, variables=None, **kwargs):
        """
        Plot the training grid samples.

        :param SolverInterface solver: The ``SolverInterface`` object.
        :param list(str) variables: Variables to plot. If None, all variables
            are plotted. If 'spatial', only spatial variables are plotted. If
            'temporal', only temporal variables are plotted. Defaults to None.

        .. todo::
            - Add support for 3D plots.
            - Fix support for more complex problems.

        :Example:
            >>> plotter = Plotter()
            >>> plotter.plot_samples(problem=problem, variables='spatial')
        """

        if variables is None:
            variables = problem.domain.variables
        elif variables == 'spatial':
            variables = problem.spatial_domain.variables
        elif variables == 'temporal':
            variables = problem.temporal_domain.variables

        if len(variables) not in [1, 2, 3]:
            raise ValueError

        fig = plt.figure()
        proj = '3d' if len(variables) == 3 else None
        ax = fig.add_subplot(projection=proj)
        for location in problem.input_pts:
            coords = problem.input_pts[location].extract(variables).T.detach()
            if coords.shape[0] == 1:  # 1D samples
                ax.plot(coords.flatten(),
                        torch.zeros(coords.flatten().shape),
                        '.',
                        label=location,
                        **kwargs)
            else:
                ax.plot(*coords, '.', label=location, **kwargs)

        ax.set_xlabel(variables[0])
        try:
            ax.set_ylabel(variables[1])
        except:
            pass

        try:
            ax.set_zlabel(variables[2])
        except:
            pass

        plt.legend()
        plt.show()

    def _1d_plot(self, pts, pred, method, truth_solution=None, **kwargs):
        """Plot solution for one dimensional function

        :param pts: Points to plot the solution.
        :type pts: torch.Tensor
        :param pred: SolverInterface solution evaluated at 'pts'.
        :type pred: torch.Tensor
        :param method: Not used, kept for code compatibility
        :type method: None
        :param truth_solution: Real solution evaluated at 'pts',
            defaults to None.
        :type truth_solution: torch.Tensor, optional
        """
        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(8, 8))

        ax.plot(pts, pred.detach(), label='Neural Network solution', **kwargs)

        if truth_solution:
            truth_output = truth_solution(pts).float()
            ax.plot(pts, truth_output.detach(), label='True solution', **kwargs)

        # TODO: pred is a torch.Tensor, so no labels is available
        #      extra variable for labels should be
        #      passed in the function arguments.
        # plt.ylabel(pred.labels[0]) 
        plt.legend()
        plt.show()

    def _2d_plot(self,
                 pts,
                 pred,
                 v,
                 res,
                 method,
                 truth_solution=None,
                 **kwargs):
        """Plot solution for two dimensional function

        :param pts: Points to plot the solution.
        :type pts: torch.Tensor
        :param pred: ``SolverInterface`` solution evaluated at 'pts'.
        :type pred: torch.Tensor
        :param method: Matplotlib method to plot 2-dimensional data,
            see https://matplotlib.org/stable/api/axes_api.html for
            reference.
        :type method: str
        :param truth_solution: Real solution evaluated at 'pts',
            defaults to None.
        :type truth_solution: torch.Tensor, optional
        """

        grids = [p_.reshape(res, res) for p_ in pts.extract(v).cpu().T]

        pred_output = pred.reshape(res, res)
        if truth_solution:
            truth_output = truth_solution(pts).float().reshape(res, res).as_subclass(torch.Tensor)
            fig, ax = plt.subplots(nrows=1, ncols=3, figsize=(16, 6))

            cb = getattr(ax[0], method)(*grids, pred_output.cpu().detach(),
                                        **kwargs)
            fig.colorbar(cb, ax=ax[0])
            ax[0].title.set_text('Neural Network prediction')
            cb = getattr(ax[1], method)(*grids, truth_output.cpu().detach(),
                                        **kwargs)
            fig.colorbar(cb, ax=ax[1])
            ax[1].title.set_text('True solution')
            cb = getattr(ax[2],
                         method)(*grids,
                                 (truth_output - pred_output).cpu().detach(),
                                 **kwargs)
            fig.colorbar(cb, ax=ax[2])
            ax[2].title.set_text('Residual')
        else:
            fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(8, 6))
            cb = getattr(ax, method)(*grids, pred_output.cpu().detach(),
                                     **kwargs)
            fig.colorbar(cb, ax=ax)
            ax.title.set_text('Neural Network prediction')

    def plot(self,
             solver,
             components=None,
             fixed_variables={},
             method='contourf',
             res=256,
             filename=None,
             **kwargs):
        """
        Plot sample of SolverInterface output.

        :param SolverInterface solver: The ``SolverInterface`` object instance.
        :param list(str) components: The output variable to plot. If None, all
            the output variables of the problem are selected. Default value is None.
        :param dict fixed_variables: A dictionary with all the variables that
            should be kept fixed during the plot. The keys of the dictionary
            are the variables name whereas the values are the corresponding
            values of the variables. Defaults is `dict()`.
        :param str method: The matplotlib method to use for
            plotting the solution. Available methods are {'contourf', 'pcolor'}.
            Default is 'contourf'.
        :param int res: The resolution, aka the number of points used for
            plotting in each axis. Default is 256.
        :param str filename: The file name to save the plot. If None, the plot
            is shown using the setted matplotlib frontend. Default is None.
        """

        if components is None:
            components = solver.problem.output_variables
        
        if len(components) > 1:
            raise NotImplementedError('Multidimensional plots are not implemented, '
                                      'set components to an available components of the problem.')
        v = [
            var for var in solver.problem.input_variables
            if var not in fixed_variables.keys()
        ]
        pts = solver.problem.domain.sample(res, 'grid', variables=v)

        fixed_pts = torch.ones(pts.shape[0], len(fixed_variables))
        fixed_pts *= torch.tensor(list(fixed_variables.values()))
        fixed_pts = fixed_pts.as_subclass(LabelTensor)
        fixed_pts.labels = list(fixed_variables.keys())

        pts = pts.append(fixed_pts)
        pts = pts.to(device=solver.device)

        predicted_output = solver.forward(pts).extract(components).as_subclass(torch.Tensor)
        truth_solution = getattr(solver.problem, 'truth_solution', None)

        if len(v) == 1:
            self._1d_plot(pts.extract(v), predicted_output, method, truth_solution,
                          **kwargs)
        elif len(v) == 2:
            self._2d_plot(pts, predicted_output, v, res, method, truth_solution,
                          **kwargs)

        plt.tight_layout()
        if filename:
            plt.savefig(filename)
        else:
            plt.show()

    def plot_loss(self,
                  trainer,
                  metrics=None,
                  logy=False,
                  logx=False,
                  filename=None,
                  **kwargs):
        """
        Plot the loss function values during traininig.

        :param trainer: the PINA Trainer object instance.
        :type trainer: Trainer
        :param str | list(str) metric: The metrics to use in the y axis. If None, the mean loss
            is plotted.
        :param bool logy: If True, the y axis is in log scale. Default is
            True.
        :param bool logx: If True, the x axis is in log scale. Default is
            True.
        :param str filename: The file name to save the plot. If None, the plot
            is shown using the setted matplotlib frontend. Default is None.
        """

        # check that MetricTracker has been used
        list_ = [
            idx for idx, s in enumerate(trainer.callbacks)
            if isinstance(s, MetricTracker)
        ]
        if not bool(list_):
            raise FileNotFoundError(
                'MetricTracker should be used as a callback during training to'
                ' use this method.')

        # extract trainer metrics
        trainer_metrics = trainer.callbacks[list_[0]].metrics
        if metrics is None:
            metrics = ['mean_loss']
        elif not isinstance(metrics, list):
            raise ValueError('metrics must be class list.')

        # loop over metrics to plot
        for metric in metrics:
            if metric not in trainer_metrics:
                raise ValueError(
                    f'{metric} not a valid metric. Available metrics are {list(trainer_metrics.keys())}.'
                )
            loss = trainer_metrics[metric]
            epochs = range(len(loss))
            plt.plot(epochs, loss, **kwargs)

        # plotting
        plt.xlabel('epoch')
        plt.ylabel('loss')
        plt.legend()

        # log axis
        if logy:
            plt.yscale('log')
        if logx:
            plt.xscale('log')

        # saving in file
        if filename:
            plt.savefig(filename)



import torch
import pytest

from pina.dataset import SamplePointDataset, SamplePointLoader, DataPointDataset
from pina import LabelTensor, Condition
from pina.equation import Equation
from pina.geometry import CartesianDomain
from pina.problem import SpatialProblem
from pina.model import FeedForward
from pina.operators import laplacian
from pina.equation.equation_factory import FixedValue


def laplace_equation(input_, output_):
    force_term = (torch.sin(input_.extract(['x'])*torch.pi) *
                    torch.sin(input_.extract(['y'])*torch.pi))
    delta_u = laplacian(output_.extract(['u']), input_)
    return delta_u - force_term

my_laplace = Equation(laplace_equation)
in_ = LabelTensor(torch.tensor([[0., 1.]]), ['x', 'y'])
out_ = LabelTensor(torch.tensor([[0.]]), ['u'])
in2_ = LabelTensor(torch.rand(60, 2), ['x', 'y'])
out2_ = LabelTensor(torch.rand(60, 1), ['u'])

class Poisson(SpatialProblem):
    output_variables = ['u']
    spatial_domain = CartesianDomain({'x': [0, 1], 'y': [0, 1]})

    conditions = {
        'gamma1': Condition(
            location=CartesianDomain({'x': [0, 1], 'y':  1}),
            equation=FixedValue(0.0)),
        'gamma2': Condition(
            location=CartesianDomain({'x': [0, 1], 'y': 0}),
            equation=FixedValue(0.0)),
        'gamma3': Condition(
            location=CartesianDomain({'x':  1, 'y': [0, 1]}),
            equation=FixedValue(0.0)),
        'gamma4': Condition(
            location=CartesianDomain({'x': 0, 'y': [0, 1]}),
            equation=FixedValue(0.0)),
        'D': Condition(
            input_points=LabelTensor(torch.rand(size=(100, 2)), ['x', 'y']),
            equation=my_laplace),
        'data': Condition(
            input_points=in_,
            output_points=out_),
        'data2': Condition(
            input_points=in2_,
            output_points=out2_)
    }

boundaries = ['gamma1', 'gamma2', 'gamma3', 'gamma4']
poisson = Poisson()
poisson.discretise_domain(10, 'grid', locations=boundaries)

def test_sample():
    sample_dataset = SamplePointDataset(poisson, device='cpu')
    assert len(sample_dataset) == 140
    assert sample_dataset.pts.shape == (140, 2)
    assert sample_dataset.pts.labels == ['x', 'y']
    assert sample_dataset.condition_indeces.dtype == torch.int64
    assert sample_dataset.condition_indeces.max() == torch.tensor(4)
    assert sample_dataset.condition_indeces.min() == torch.tensor(0)

def test_data():
    dataset = DataPointDataset(poisson, device='cpu')
    assert len(dataset) == 61
    assert dataset.input_pts.shape == (61, 2)
    assert dataset.input_pts.labels == ['x', 'y']
    assert dataset.output_pts.shape == (61, 1 )
    assert dataset.output_pts.labels == ['u']
    assert dataset.condition_indeces.dtype == torch.int64
    assert dataset.condition_indeces.max() == torch.tensor(1)
    assert dataset.condition_indeces.min() == torch.tensor(0)

def test_loader():
    sample_dataset = SamplePointDataset(poisson, device='cpu')
    data_dataset = DataPointDataset(poisson, device='cpu')
    loader = SamplePointLoader(sample_dataset, data_dataset, batch_size=10)

    for batch in loader:
        assert len(batch) in [2, 3]
        assert batch['pts'].shape[0] <= 10
        assert batch['pts'].requires_grad == True
        assert batch['pts'].labels == ['x', 'y']

    loader2 = SamplePointLoader(sample_dataset, data_dataset, batch_size=None)
    assert len(list(loader2)) == 2

def test_loader2():
    poisson2 = Poisson()
    del poisson.conditions['data2']
    del poisson2.conditions['data']
    poisson2.discretise_domain(10, 'grid', locations=boundaries)
    sample_dataset = SamplePointDataset(poisson, device='cpu')
    data_dataset = DataPointDataset(poisson, device='cpu')
    loader = SamplePointLoader(sample_dataset, data_dataset, batch_size=10)

    for batch in loader:
        assert len(batch) == 2 # only phys condtions
        assert batch['pts'].shape[0] <= 10
        assert batch['pts'].requires_grad == True
        assert batch['pts'].labels == ['x', 'y']

def test_loader3():
    poisson2 = Poisson()
    del poisson.conditions['gamma1']
    del poisson.conditions['gamma2']
    del poisson.conditions['gamma3']
    del poisson.conditions['gamma4']
    del poisson.conditions['D']
    sample_dataset = SamplePointDataset(poisson, device='cpu')
    data_dataset = DataPointDataset(poisson, device='cpu')
    loader = SamplePointLoader(sample_dataset, data_dataset, batch_size=10)

    for batch in loader:
        assert len(batch) == 2 # only phys condtions
        assert batch['pts'].shape[0] <= 10
        assert batch['pts'].requires_grad == True
        assert batch['pts'].labels == ['x', 'y']



__all__ = ['SwitchOptimizer', 'R3Refinement', 'MetricTracker']

from .optimizer_callbacks import SwitchOptimizer
from .adaptive_refinment_callbacks import R3Refinement
from .processing_callbacks import MetricTracker



from setuptools import setup, find_packages

meta = {}
with open("pina/meta.py") as fp:
    exec(fp.read(), meta)

# Package meta-data.
IMPORTNAME = meta['__title__']
PIPNAME = meta['__packagename__']
DESCRIPTION = 'Physic Informed Neural networks for Advance modeling.'
URL = 'https://github.com/mathLab/PINA'
MAIL = meta['__mail__']
AUTHOR = meta['__author__']
VERSION = meta['__version__']
KEYWORDS = 'physics-informed neural-network'

REQUIRED = [
    'numpy', 'matplotlib', 'torch', 'lightning', 'pytorch_lightning'
]

EXTRAS = {
    'docs': ['sphinx', 'sphinx_rtd_theme'],
    'test': ['pytest', 'pytest-cov'],
}

LDESCRIPTION = (
    "PINA is a Python package providing an easy interface to deal with "
    "physics-informed neural networks (PINN) for the approximation of "
    "(differential, nonlinear, ...) functions. Based on Pytorch, PINA "
    "offers a simple and intuitive way to formalize a specific problem "
    "and solve it using PINN. The approximated solution of a differential "
    "equation can be implemented using PINA in a few lines of code thanks "
    "to the intuitive and user-friendly interface."
)

setup(
    name=PIPNAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LDESCRIPTION,
    author=AUTHOR,
    author_email=MAIL,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Mathematics'
    ],
    keywords=KEYWORDS,
    url=URL,
    license='MIT',
    packages=find_packages(),
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    include_package_data=True,
    zip_safe=False,
)



__all__ = [
    'PINN', 'Trainer', 'LabelTensor', 'Plotter', 'Condition',
    'SamplePointDataset', 'SamplePointLoader'
]

from .meta import *
from .label_tensor import LabelTensor
from .solvers.solver import SolverInterface
from .trainer import Trainer
from .plotter import Plotter
from .condition import Condition
from .dataset import SamplePointDataset
from .dataset import SamplePointLoader



"""Module for Location class."""

from abc import ABCMeta, abstractmethod


class Location(metaclass=ABCMeta):
    """
    Abstract Location class.
    Any geometry entity should inherit from this class.
    """

    @abstractmethod
    def sample(self):
        """
        Abstract method for sampling a point from the location. To be
        implemented in the child class.
        """
        pass

    @abstractmethod
    def is_inside(self, point, check_border=False):
        """
        Abstract method for checking if a point is inside the location. To be
        implemented in the child class.

        :param torch.Tensor point: A tensor point to be checked.
        :param bool check_border: A boolean that determines whether the border
            of the location is considered checked to be considered inside or
            not. Defaults to ``False``.
        """
        pass



import torch
import torch.nn as nn
from ...utils import check_consistency

from pina.model.layers import SpectralConvBlock1D, SpectralConvBlock2D, SpectralConvBlock3D


class FourierBlock1D(nn.Module):
    """
    Fourier block implementation for three dimensional
    input tensor. The combination of Fourier blocks
    make up the Fourier Neural Operator  

    .. seealso::

        **Original reference**: Li, Z., Kovachki, N., Azizzadenesheli, K., Liu, B.,
        Bhattacharya, K., Stuart, A., & Anandkumar, A. (2020). *Fourier neural operator for
        parametric partial differential equations*.
        DOI: `arXiv preprint arXiv:2010.08895.
        <https://arxiv.org/abs/2010.08895>`_

    """

    def __init__(self,
                 input_numb_fields,
                 output_numb_fields,
                 n_modes,
                 activation=torch.nn.Tanh):
        super().__init__()
        """
        PINA implementation of Fourier block one dimension. The module computes
        the spectral convolution of the input with a linear kernel in the
        fourier space, and then it maps the input back to the physical
        space. The output is then added to a Linear tranformation of the
        input in the physical space. Finally an activation function is
        applied to the output. 

        The block expects an input of size ``[batch, input_numb_fields, N]``
        and returns an output of size ``[batch, output_numb_fields, N]``.

        :param int input_numb_fields: The number of channels for the input.
        :param int output_numb_fields: The number of channels for the output.
        :param list | tuple n_modes: Number of modes to select for each dimension.
            It must be at most equal to the ``floor(N/2)+1``.
        :param torch.nn.Module activation: The activation function.
        """
        # check type consistency
        check_consistency(activation(), nn.Module)

        # assign variables
        self._spectral_conv = SpectralConvBlock1D(
            input_numb_fields=input_numb_fields,
            output_numb_fields=output_numb_fields,
            n_modes=n_modes)
        self._activation = activation()
        self._linear = nn.Conv1d(input_numb_fields, output_numb_fields, 1)

    def forward(self, x):
        """
        Forward computation for Fourier Block. It performs a spectral 
        convolution and a linear transformation of the input and sum the
        results.

        :param x: The input tensor for fourier block, expect of size 
            ``[batch, input_numb_fields, x]``.
        :type x: torch.Tensor
        :return: The output tensor obtained from the
            fourier block of size ``[batch, output_numb_fields, x]``.
        :rtype: torch.Tensor
        """
        return self._activation(self._spectral_conv(x) + self._linear(x))


class FourierBlock2D(nn.Module):
    """
    Fourier block implementation for two dimensional
    input tensor. The combination of Fourier blocks
    make up the Fourier Neural Operator    

    .. seealso::

        **Original reference**: Li, Zongyi, et al.
        *Fourier neural operator for parametric partial
        differential equations*. arXiv preprint
        arXiv:2010.08895 (2020)
        <https://arxiv.org/abs/2010.08895.pdf>`_.

    """

    def __init__(self,
                 input_numb_fields,
                 output_numb_fields,
                 n_modes,
                 activation=torch.nn.Tanh):
        """
        PINA implementation of Fourier block two dimensions. The module computes
        the spectral convolution of the input with a linear kernel in the
        fourier space, and then it maps the input back to the physical
        space. The output is then added to a Linear tranformation of the
        input in the physical space. Finally an activation function is
        applied to the output. 

        The block expects an input of size ``[batch, input_numb_fields, Nx, Ny]``
        and returns an output of size ``[batch, output_numb_fields, Nx, Ny]``.

        :param int input_numb_fields: The number of channels for the input.
        :param int output_numb_fields: The number of channels for the output.
        :param list | tuple n_modes: Number of modes to select for each dimension.
            It must be at most equal to the ``floor(Nx/2)+1`` and ``floor(Ny/2)+1``.
        :param torch.nn.Module activation: The activation function.
        """
        super().__init__()

        # check type consistency
        check_consistency(activation(), nn.Module)

        # assign variables
        self._spectral_conv = SpectralConvBlock2D(
            input_numb_fields=input_numb_fields,
            output_numb_fields=output_numb_fields,
            n_modes=n_modes)
        self._activation = activation()
        self._linear = nn.Conv2d(input_numb_fields, output_numb_fields, 1)

    def forward(self, x):
        """
        Forward computation for Fourier Block. It performs a spectral 
        convolution and a linear transformation of the input and sum the
        results.

        :param x: The input tensor for fourier block, expect of size 
            ``[batch, input_numb_fields, x, y]``.
        :type x: torch.Tensor
        :return: The output tensor obtained from the
            fourier block of size ``[batch, output_numb_fields, x, y, z]``.
        :rtype: torch.Tensor
        """
        return self._activation(self._spectral_conv(x) + self._linear(x))


class FourierBlock3D(nn.Module):
    """
    Fourier block implementation for three dimensional
    input tensor. The combination of Fourier blocks
    make up the Fourier Neural Operator  

    .. seealso::

        **Original reference**: Li, Zongyi, et al.
        *Fourier neural operator for parametric partial
        differential equations*. arXiv preprint
        arXiv:2010.08895 (2020)
        <https://arxiv.org/abs/2010.08895.pdf>`_.

    """

    def __init__(self,
                 input_numb_fields,
                 output_numb_fields,
                 n_modes,
                 activation=torch.nn.Tanh):
        """
        PINA implementation of Fourier block three dimensions. The module computes
        the spectral convolution of the input with a linear kernel in the
        fourier space, and then it maps the input back to the physical
        space. The output is then added to a Linear tranformation of the
        input in the physical space. Finally an activation function is
        applied to the output. 

        The block expects an input of size ``[batch, input_numb_fields, Nx, Ny, Nz]``
        and returns an output of size ``[batch, output_numb_fields, Nx, Ny, Nz]``.

        :param int input_numb_fields: The number of channels for the input.
        :param int output_numb_fields: The number of channels for the output.
        :param list | tuple n_modes: Number of modes to select for each dimension.
            It must be at most equal to the ``floor(Nx/2)+1``, ``floor(Ny/2)+1``
            and ``floor(Nz/2)+1``.
        :param torch.nn.Module activation: The activation function.
        """
        super().__init__()

        # check type consistency
        check_consistency(activation(), nn.Module)

        # assign variables
        self._spectral_conv = SpectralConvBlock3D(
            input_numb_fields=input_numb_fields,
            output_numb_fields=output_numb_fields,
            n_modes=n_modes)
        self._activation = activation()
        self._linear = nn.Conv3d(input_numb_fields, output_numb_fields, 1)

    def forward(self, x):
        """
        Forward computation for Fourier Block. It performs a spectral 
        convolution and a linear transformation of the input and sum the
        results.

        :param x: The input tensor for fourier block, expect of size 
            ``[batch, input_numb_fields, x, y, z]``.
        :type x: torch.Tensor
        :return: The output tensor obtained from the
            fourier block of size ``[batch, output_numb_fields, x, y, z]``.
        :rtype: torch.Tensor
        """
        return self._activation(self._spectral_conv(x) + self._linear(x))



__all__ = [
    '__project__', '__title__', '__author__', '__copyright__', '__license__',
    '__version__', '__mail__', '__maintainer__', '__status__'
]

__project__ = 'PINA'
__title__ = "pina"
__author__ = "PINA Contributors"
__copyright__ = "Copyright 2021-2023, PINA Contributors"
__license__ = "MIT"
__version__ = "0.1"
__mail__ = 'demo.nicola@gmail.com, dario.coscia@sissa.it'  # TODO
__maintainer__ = __author__
__status__ = "Alpha"
__packagename__ = "pina-mathlab"



import torch
from torch.nn.parameter import Parameter


class AdaptiveTanh(torch.nn.Module):
    '''
    Implementation of soft exponential activation.
    Shape:
        - Input: (N, *) where * means, any number of additional
          dimensions
        - Output: (N, *), same shape as the input
    Parameters:
        - alpha - trainable parameter
    References:
        - See related paper:
        https://arxiv.org/pdf/1602.01321.pdf
    Examples:
        >>> a1 = soft_exponential(256)
        >>> x = torch.randn(256)
        >>> x = a1(x)
    '''

    def __init__(self, alpha=None):
        '''
        Initialization.
        INPUT:
            - in_features: shape of the input
            - aplha: trainable parameter
            aplha is initialized with zero value by default
        '''
        super(AdaptiveTanh, self).__init__()
        #self.in_features = in_features

        # initialize alpha
        if alpha == None:
            self.alpha = Parameter(
                torch.tensor(1.0))  # create a tensor out of alpha
        else:
            self.alpha = Parameter(
                torch.tensor(alpha))  # create a tensor out of alpha

        self.alpha.requiresGrad = True  # set requiresGrad to true!

        self.scale = Parameter(torch.tensor(1.0))
        self.scale.requiresGrad = True  # set requiresGrad to true!

        self.translate = Parameter(torch.tensor(0.0))
        self.translate.requiresGrad = True  # set requiresGrad to true!

    def forward(self, x):
        '''
        Forward pass of the function.
        Applies the function to the input elementwise.
        '''
        x += self.translate
        return self.scale * (torch.exp(self.alpha * x) - torch.exp(
            -self.alpha * x)) / (torch.exp(self.alpha * x) +
                                 torch.exp(-self.alpha * x))



__all__ = [
    'PINN',
    'GAROM',
    'SupervisedSolver',
    'SolverInterface'

]

from .garom import GAROM
from .pinn import PINN
from .supervised import SupervisedSolver
from .solver import SolverInterface



import torch
import pytest

from pina.problem import SpatialProblem
from pina.operators import laplacian
from pina import LabelTensor, Condition
from pina.geometry import CartesianDomain
from pina.equation.equation import Equation
from pina.equation.equation_factory import FixedValue


def laplace_equation(input_, output_):
    force_term = (torch.sin(input_.extract(['x']) * torch.pi) *
                  torch.sin(input_.extract(['y']) * torch.pi))
    delta_u = laplacian(output_.extract(['u']), input_)
    return delta_u - force_term


my_laplace = Equation(laplace_equation)
in_ = LabelTensor(torch.tensor([[0., 1.]], requires_grad=True), ['x', 'y'])
out_ = LabelTensor(torch.tensor([[0.]], requires_grad=True), ['u'])


class Poisson(SpatialProblem):
    output_variables = ['u']
    spatial_domain = CartesianDomain({'x': [0, 1], 'y': [0, 1]})

    conditions = {
        'gamma1':
        Condition(location=CartesianDomain({
            'x': [0, 1],
            'y': 1
        }),
                  equation=FixedValue(0.0)),
        'gamma2':
        Condition(location=CartesianDomain({
            'x': [0, 1],
            'y': 0
        }),
                  equation=FixedValue(0.0)),
        'gamma3':
        Condition(location=CartesianDomain({
            'x': 1,
            'y': [0, 1]
        }),
                  equation=FixedValue(0.0)),
        'gamma4':
        Condition(location=CartesianDomain({
            'x': 0,
            'y': [0, 1]
        }),
                  equation=FixedValue(0.0)),
        'D':
        Condition(location=CartesianDomain({
            'x': [0, 1],
            'y': [0, 1]
        }),
                  equation=my_laplace),
        'data':
        Condition(input_points=in_, output_points=out_)
    }

    def poisson_sol(self, pts):
        return -(torch.sin(pts.extract(['x']) * torch.pi) *
                 torch.sin(pts.extract(['y']) * torch.pi)) / (2 * torch.pi**2)

    truth_solution = poisson_sol


# make the problem
poisson_problem = Poisson()


def test_discretise_domain():
    n = 10
    boundaries = ['gamma1', 'gamma2', 'gamma3', 'gamma4']
    poisson_problem.discretise_domain(n, 'grid', locations=boundaries)
    for b in boundaries:
        assert poisson_problem.input_pts[b].shape[0] == n
    poisson_problem.discretise_domain(n, 'random', locations=boundaries)
    for b in boundaries:
        assert poisson_problem.input_pts[b].shape[0] == n

    poisson_problem.discretise_domain(n, 'grid', locations=['D'])
    assert poisson_problem.input_pts['D'].shape[0] == n**2
    poisson_problem.discretise_domain(n, 'random', locations=['D'])
    assert poisson_problem.input_pts['D'].shape[0] == n

    poisson_problem.discretise_domain(n, 'latin', locations=['D'])
    assert poisson_problem.input_pts['D'].shape[0] == n

    poisson_problem.discretise_domain(n, 'lh', locations=['D'])
    assert poisson_problem.input_pts['D'].shape[0] == n


def test_sampling_few_variables():
    n = 10
    poisson_problem.discretise_domain(n,
                                      'grid',
                                      locations=['D'],
                                      variables=['x'])
    assert poisson_problem.input_pts['D'].shape[1] == 1
    assert poisson_problem._have_sampled_points['D'] is False


# def test_sampling_all_args():
#     n = 10
#     poisson_problem.discretise_domain(n, 'grid', locations=['D'])

# def test_sampling_all_kwargs():
#     n = 10
#     poisson_problem.discretise_domain(n=n, mode='latin', locations=['D'])

# def test_sampling_dict():
#     n = 10
#     poisson_problem.discretise_domain(
#         {'variables': ['x', 'y'], 'mode': 'grid', 'n': n}, locations=['D'])

# def test_sampling_mixed_args_kwargs():
#     n = 10
#     with pytest.raises(ValueError):
#         poisson_problem.discretise_domain(n, mode='latin', locations=['D'])



"""Module for FeedForward model"""
import torch
import torch.nn as nn
from ..utils import check_consistency
from .layers.residual import EnhancedLinear


class FeedForward(torch.nn.Module):
    """
    The PINA implementation of feedforward network, also refered as multilayer
    perceptron.

    :param int input_dimensions: The number of input components of the model.
        Expected tensor shape of the form :math:`(*, d)`, where *
        means any number of dimensions including none, and :math:`d` the ``input_dimensions``.
    :param int output_dimensions: The number of output components of the model.
        Expected tensor shape of the form :math:`(*, d)`, where *
        means any number of dimensions including none, and :math:`d` the ``output_dimensions``.
    :param int inner_size: number of neurons in the hidden layer(s). Default is
        20.
    :param int n_layers: number of hidden layers. Default is 2.
    :param func: the activation function to use. If a single
        :class:`torch.nn.Module` is passed, this is used as activation function
        after any layers, except the last one. If a list of Modules is passed,
        they are used as activation functions at any layers, in order.
    :param list(int) | tuple(int) layers: a list containing the number of neurons for
        any hidden layers. If specified, the parameters ``n_layers`` e
        ``inner_size`` are not considered.
    :param bool bias: If ``True`` the MLP will consider some bias.
    """

    def __init__(self,
                 input_dimensions,
                 output_dimensions,
                 inner_size=20,
                 n_layers=2,
                 func=nn.Tanh,
                 layers=None,
                 bias=True):
        """
        """
        super().__init__()

        if not isinstance(input_dimensions, int):
            raise ValueError('input_dimensions expected to be int.')
        self.input_dimension = input_dimensions

        if not isinstance(output_dimensions, int):
            raise ValueError('output_dimensions expected to be int.')
        self.output_dimension = output_dimensions
        if layers is None:
            layers = [inner_size] * n_layers

        tmp_layers = layers.copy()
        tmp_layers.insert(0, self.input_dimension)
        tmp_layers.append(self.output_dimension)

        self.layers = []
        for i in range(len(tmp_layers) - 1):
            self.layers.append(
                nn.Linear(tmp_layers[i], tmp_layers[i + 1], bias=bias))

        if isinstance(func, list):
            self.functions = func
        else:
            self.functions = [func for _ in range(len(self.layers) - 1)]

        if len(self.layers) != len(self.functions) + 1:
            raise RuntimeError('uncosistent number of layers and functions')

        unique_list = []
        for layer, func in zip(self.layers[:-1], self.functions):
            unique_list.append(layer)
            if func is not None:
                unique_list.append(func())
        unique_list.append(self.layers[-1])

        self.model = nn.Sequential(*unique_list)

    def forward(self, x):
        """
        Defines the computation performed at every call.

        :param x: The tensor to apply the forward pass.
        :type x: torch.Tensor
        :return: the output computed by the model.
        :rtype: torch.Tensor
        """
        return self.model(x)


class ResidualFeedForward(torch.nn.Module):
    """
    The PINA implementation of feedforward network, also with skipped connection
    and transformer network, as presented in **Understanding and mitigating gradient
    pathologies in physics-informed neural networks**

    .. seealso::

        **Original reference**: Wang, Sifan, Yujun Teng, and Paris Perdikaris. 
        *Understanding and mitigating gradient flow pathologies in physics-informed
        neural networks*. SIAM Journal on Scientific Computing 43.5 (2021): A3055-A3081.
        DOI: `10.1137/20M1318043
        <https://epubs.siam.org/doi/abs/10.1137/20M1318043>`_


    :param int input_dimensions: The number of input components of the model.
        Expected tensor shape of the form :math:`(*, d)`, where *
        means any number of dimensions including none, and :math:`d` the ``input_dimensions``.
    :param int output_dimensions: The number of output components of the model.
        Expected tensor shape of the form :math:`(*, d)`, where *
        means any number of dimensions including none, and :math:`d` the ``output_dimensions``.
    :param int inner_size: number of neurons in the hidden layer(s). Default is
        20.
    :param int n_layers: number of hidden layers. Default is 2.
    :param func: the activation function to use. If a single
        :class:`torch.nn.Module` is passed, this is used as activation function
        after any layers, except the last one. If a list of Modules is passed,
        they are used as activation functions at any layers, in order.
    :param bool bias: If ``True`` the MLP will consider some bias.
    :param list | tuple transformer_nets: a list or tuple containing the two
        torch.nn.Module which act as transformer network. The input dimension
        of the network must be the same as ``input_dimensions``, and the output
        dimension must be the same as ``inner_size``.
    """

    def __init__(self,
                 input_dimensions,
                 output_dimensions,
                 inner_size=20,
                 n_layers=2,
                 func=nn.Tanh,
                 bias=True,
                 transformer_nets=None):
        """
        """
        super().__init__()

        # check type consistency
        check_consistency(input_dimensions, int)
        check_consistency(output_dimensions, int)
        check_consistency(inner_size, int)
        check_consistency(n_layers, int)
        check_consistency(func, torch.nn.Module, subclass=True)
        check_consistency(bias, bool)

        # check transformer nets
        if transformer_nets is None:
            transformer_nets = [
                EnhancedLinear(
                    nn.Linear(in_features=input_dimensions,
                              out_features=inner_size), nn.Tanh()),
                EnhancedLinear(
                    nn.Linear(in_features=input_dimensions,
                              out_features=inner_size), nn.Tanh())
            ]
        elif isinstance(transformer_nets, (list, tuple)):
            if len(transformer_nets) != 2:
                raise ValueError(
                    'transformer_nets needs to be a list of len two.')
            for net in transformer_nets:
                if not isinstance(net, nn.Module):
                    raise ValueError(
                        'transformer_nets needs to be a list of torch.nn.Module.'
                    )
                x = torch.rand(10, input_dimensions)
                try:
                    out = net(x)
                except RuntimeError:
                    raise ValueError(
                        'transformer network input incompatible with input_dimensions.'
                    )
                if out.shape[-1] != inner_size:
                    raise ValueError(
                        'transformer network output incompatible with inner_size.'
                    )
        else:
            RuntimeError(
                'Runtime error for transformer nets, check official documentation.'
            )

        # assign variables
        self.input_dimension = input_dimensions
        self.output_dimension = output_dimensions
        self.transformer_nets = nn.ModuleList(transformer_nets)

        # build layers
        layers = [inner_size] * n_layers

        tmp_layers = layers.copy()
        tmp_layers.insert(0, self.input_dimension)

        self.layers = []
        for i in range(len(tmp_layers) - 1):
            self.layers.append(
                nn.Linear(tmp_layers[i], tmp_layers[i + 1], bias=bias))
        self.last_layer = nn.Linear(tmp_layers[len(tmp_layers) - 1],
                                    output_dimensions,
                                    bias=bias)

        if isinstance(func, list):
            self.functions = func()
        else:
            self.functions = [func() for _ in range(len(self.layers))]

        if len(self.layers) != len(self.functions):
            raise RuntimeError('uncosistent number of layers and functions')

        unique_list = []
        for layer, func in zip(self.layers, self.functions):
            unique_list.append(EnhancedLinear(layer=layer, activation=func))
        self.inner_layers = torch.nn.Sequential(*unique_list)

    def forward(self, x):
        """
        Defines the computation performed at every call.

        :param x: The tensor to apply the forward pass.
        :type x: torch.Tensor
        :return: the output computed by the model.
        :rtype: torch.Tensor
        """
        # enhance the input with transformer
        input_ = []
        for nets in self.transformer_nets:
            input_.append(nets(x))

        # skip connections pass
        for layer in self.inner_layers.children():
            x = layer(x)
            x = (1. - x) * input_[0] + x * input_[1]

        # last layer
        return self.last_layer(x)



""" Solver module. """

from abc import ABCMeta, abstractmethod
from ..model.network import Network
import pytorch_lightning
from ..utils import check_consistency
from ..problem import AbstractProblem
import torch


class SolverInterface(pytorch_lightning.LightningModule, metaclass=ABCMeta):
    """
    Solver base class. This class inherits is a wrapper of
    LightningModule class, inheriting all the
    LightningModule methods.
    """

    def __init__(self,
                 models,
                 problem,
                 optimizers,
                 optimizers_kwargs,
                 extra_features=None):
        """
        :param models: A torch neural network model instance.
        :type models: torch.nn.Module
        :param problem: A problem definition instance.
        :type problem: AbstractProblem
        :param list(torch.optim.Optimizer) optimizer: A list of neural network optimizers to
            use.
        :param list(dict) optimizer_kwargs: A list of optimizer constructor keyword args.
        :param list(torch.nn.Module) extra_features: The additional input
            features to use as augmented input. If ``None`` no extra features 
            are passed. If it is a list of :class:`torch.nn.Module`, the extra feature
            list is passed to all models. If it is a list of extra features' lists,
            each single list of extra feature is passed to a model.
        """
        super().__init__()

        # check consistency of the inputs
        check_consistency(models, torch.nn.Module)
        check_consistency(problem, AbstractProblem)
        check_consistency(optimizers, torch.optim.Optimizer, subclass=True)
        check_consistency(optimizers_kwargs, dict)

        # put everything in a list if only one input
        if not isinstance(models, list):
            models = [models]
        if not isinstance(optimizers, list):
            optimizers = [optimizers]
            optimizers_kwargs = [optimizers_kwargs]

        # number of models and optimizers
        len_model = len(models)
        len_optimizer = len(optimizers)
        len_optimizer_kwargs = len(optimizers_kwargs)

        # check length consistency optimizers
        if len_model != len_optimizer:
            raise ValueError('You must define one optimizer for each model.'
                             f'Got {len_model} models, and {len_optimizer}'
                             ' optimizers.')

        # check length consistency optimizers kwargs
        if len_optimizer_kwargs != len_optimizer:
            raise ValueError('You must define one dictionary of keyword'
                             ' arguments for each optimizers.'
                             f'Got {len_optimizer} optimizers, and'
                             f' {len_optimizer_kwargs} dicitionaries')

        # extra features handling
        if (extra_features is None) or (len(extra_features)==0):
            extra_features = [None] * len_model
        else:
            # if we only have a list of extra features
            if not isinstance(extra_features[0], (tuple, list)):
                extra_features = [extra_features] * len_model
            else:  # if we have a list of list extra features
                if len(extra_features) != len_model:
                    raise ValueError(
                        'You passed a list of extrafeatures list with len'
                        f'different of models len. Expected {len_model} '
                        f'got {len(extra_features)}. If you want to use '
                        'the same list of extra features for all models, '
                        'just pass a list of extrafeatures and not a list '
                        'of list of extra features.')

        # assigning model and optimizers
        self._pina_models = []
        self._pina_optimizers = []

        for idx in range(len_model):
            model_ = Network(model=models[idx],
                             input_variables=problem.input_variables,
                             output_variables=problem.output_variables,
                             extra_features=extra_features[idx])
            optim_ = optimizers[idx](model_.parameters(),
                                     **optimizers_kwargs[idx])
            self._pina_models.append(model_)
            self._pina_optimizers.append(optim_)

        # assigning problem
        self._pina_problem = problem

    @abstractmethod
    def forward(self, *args, **kwargs):
        pass

    @abstractmethod
    def training_step(self):
        pass

    @abstractmethod
    def configure_optimizers(self):
        pass

    @property
    def models(self):
        """
        The torch model."""
        return self._pina_models

    @property
    def optimizers(self):
        """
        The torch model."""
        return self._pina_optimizers

    @property
    def problem(self):
        """
        The problem formulation."""
        return self._pina_problem

    # @model.setter
    # def model(self, new_model):
    #     """
    #     Set the torch."""
    #     check_consistency(new_model, nn.Module, 'torch model')
    #     self._model= new_model

    # @problem.setter
    # def problem(self, problem):
    #     """
    #     Set the problem formulation."""
    #     check_consistency(problem, AbstractProblem, 'pina problem')
    #     self._problem = problem



""" Module for SystemEquation. """
import torch
from .equation import Equation
from ..utils import check_consistency


class SystemEquation(Equation):

    def __init__(self, list_equation, reduction='mean'):
        """
        System of Equation class for specifing any system
        of equations in PINA.
        Each ``equation`` passed to a ``Condition`` object
        must be an ``Equation`` or ``SystemEquation``.
        A ``SystemEquation`` is specified by a list of
        equations.

        :param Callable equation: A ``torch`` callable equation to
            evaluate the residual
        :param str reduction: Specifies the reduction to apply to the output:
            ``none`` | ``mean`` | ``sum``  | ``callable``. ``none``: no reduction
            will be applied, ``mean``: the sum of the output will be divided
            by the number of elements in the output, ``sum``: the output will
            be summed. ``callable`` a callable function to perform reduction,
            no checks guaranteed. Default: ``mean``.
        """
        check_consistency([list_equation], list)
        check_consistency(reduction, str)

        # equations definition
        self.equations = []
        for _, equation in enumerate(list_equation):
            self.equations.append(Equation(equation))

        # possible reduction
        if reduction == 'mean':
            self.reduction = torch.mean
        elif reduction == 'sum':
            self.reduction = torch.sum
        elif (reduction == 'none') or callable(reduction):
            self.reduction = reduction
        else:
            raise NotImplementedError(
                'Only mean and sum reductions implemented.')

    def residual(self, input_, output_, params_=None):
        """
        Residual computation for the equations of the system.

        :param LabelTensor input_: Input points to evaluate the system of
            equations.
        :param LabelTensor output_: Output vectors given by a model (e.g,
            a ``FeedForward`` model).
        :param dict params_: Dictionary of parameters related to the inverse
            problem (if any).
            If the equation is not related to an ``InverseProblem``, the
            parameters are initialized to ``None`` and the residual is
            computed as ``equation(input_, output_)``.
            Otherwise, the parameters are automatically initialized in the
            ranges specified by the user.

        :return: The residual evaluation of the specified system of equations,
            aggregated by the ``reduction`` defined in the ``__init__``.
        :rtype: LabelTensor
        """
        residual = torch.hstack(
            [equation.residual(input_, output_, params_) for equation in self.equations])

        if self.reduction == 'none':
            return residual

        return self.reduction(residual, dim=-1)
